from __future__ import annotations

from cognite.client.data_classes._base import T_CogniteResourceList, T_WritableCogniteResource, T_WriteClass
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.loaders import (
    ResourceLoader,
)
from cognite_toolkit._cdf_tk.loaders._base_loaders import T_ID, T_WritableCogniteResourceList


def _print_ids_or_length(resource_ids: SequenceNotStr[T_ID], limit: int = 10) -> str:
    if len(resource_ids) == 1:
        return f"{resource_ids[0]!r}"
    elif len(resource_ids) <= limit:
        return f"{resource_ids}"
    else:
        return f"{len(resource_ids)} items"


def _remove_duplicates(
    loaded_resources: T_CogniteResourceList,
    loader: ResourceLoader[
        T_ID, T_WriteClass, T_WritableCogniteResource, T_CogniteResourceList, T_WritableCogniteResourceList
    ],
) -> tuple[T_CogniteResourceList, list[T_ID]]:
    seen: set[T_ID] = set()
    output = loader.list_write_cls([])
    duplicates: list[T_ID] = []
    for item in loaded_resources:
        identifier = loader.get_id(item)
        if identifier not in seen:
            output.append(item)
            seen.add(identifier)
        else:
            duplicates.append(identifier)
    return output, duplicates



import collections.abc
from collections.abc import Hashable, Iterable
from functools import lru_cache
from pathlib import Path
from typing import Any, final

from cognite.client.data_classes.capabilities import Capability, DataModelInstancesAcl
from cognite.client.data_classes.data_modeling import NodeApplyResultList, NodeId
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk._parameters import ParameterSpec, ParameterSpecSet
from cognite_toolkit._cdf_tk.client.data_classes.apm_config_v1 import (
    APMConfig,
    APMConfigList,
    APMConfigWrite,
    APMConfigWriteList,
)
from cognite_toolkit._cdf_tk.constants import BUILD_FOLDER_ENCODING
from cognite_toolkit._cdf_tk.loaders._base_loaders import ResourceLoader
from cognite_toolkit._cdf_tk.utils import quote_int_value_by_key_in_yaml, safe_read
from cognite_toolkit._cdf_tk.utils.cdf import iterate_instances
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_hashable, diff_list_identifiable, hash_dict

from .auth_loaders import GroupAllScopedLoader
from .classic_loaders import AssetLoader
from .data_organization_loaders import DataSetsLoader
from .datamodel_loaders import SpaceLoader
from .group_scoped_loader import GroupResourceScopedLoader


@final
class InfieldV1Loader(ResourceLoader[str, APMConfigWrite, APMConfig, APMConfigWriteList, APMConfigList]):
    folder_name = "cdf_applications"
    filename_pattern = r"^.*\.InfieldV1$"  # Matches all yaml files whose stem ends with '.InfieldV1'.
    filetypes = frozenset({"yaml", "yml"})
    resource_cls = APMConfig
    resource_write_cls = APMConfigWrite
    list_cls = APMConfigList
    list_write_cls = APMConfigWriteList
    kind = "InfieldV1"
    dependencies = frozenset(
        {DataSetsLoader, AssetLoader, SpaceLoader, GroupAllScopedLoader, GroupResourceScopedLoader}
    )
    _doc_url = "Instances/operation/applyNodeAndEdges"
    _root_location_filters: tuple[str, ...] = ("general", "assets", "files", "timeseries")
    _group_keys: tuple[str, ...] = ("templateAdmins", "checklistAdmins")

    @property
    def display_name(self) -> str:
        return "infield configs"

    @classmethod
    def get_id(cls, item: APMConfig | APMConfigWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if not item.external_id:
            raise KeyError("APMConfig must have external_id")
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: collections.abc.Sequence[APMConfigWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [DataModelInstancesAcl.Action.Read]
            if read_only
            else [DataModelInstancesAcl.Action.Read, DataModelInstancesAcl.Action.Write]
        )

        return DataModelInstancesAcl(actions, DataModelInstancesAcl.Scope.SpaceID([APMConfig.space]))

    def create(self, items: APMConfigWriteList) -> NodeApplyResultList:
        result = self.client.data_modeling.instances.apply(
            nodes=items.as_nodes(), auto_create_direct_relations=True, replace=False
        )
        return result.nodes

    def retrieve(self, ids: SequenceNotStr[str]) -> APMConfigList:
        result = self.client.data_modeling.instances.retrieve(
            nodes=self._as_node_ids(ids), sources=APMConfig.view_id
        ).nodes
        return APMConfigList.from_nodes(result)

    def update(self, items: APMConfigWriteList) -> NodeApplyResultList:
        result = self.client.data_modeling.instances.apply(
            nodes=items.as_nodes(), auto_create_direct_relations=True, replace=True
        )
        return result.nodes

    def delete(self, ids: SequenceNotStr[str]) -> int:
        try:
            deleted = self.client.data_modeling.instances.delete(nodes=self._as_node_ids(ids))
        except CogniteAPIError as e:
            if "not exist" in e.message and "space" in e.message.lower():
                return 0
            raise e
        return len(deleted.nodes)

    @staticmethod
    def _as_node_ids(ids: SequenceNotStr[str]) -> list[NodeId]:
        return [NodeId(APMConfig.space, id) for id in ids]

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[APMConfig]:
        for node in iterate_instances(
            self.client, space=space, instance_type="node", source=APMConfig.view_id, console=self.console
        ):
            yield APMConfig.from_node(node)

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        spec = super().get_write_cls_parameter_spec()
        # Added by toolkit
        spec.add(
            ParameterSpec(
                ("featureConfiguration", "rootLocationConfiguration", "dataSetExternalId"),
                frozenset({"str"}),
                is_required=False,
                _is_nullable=False,
            )
        )
        spec.discard(
            ParameterSpec(
                (
                    "featureConfiguration",
                    "rootLocationConfiguration",
                    "dataSetId",
                ),
                frozenset({"int"}),
                is_required=False,
                _is_nullable=False,
            )
        )
        return spec

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        if isinstance(app_data_space_id := item.get("appDataSpaceId"), str):
            yield SpaceLoader, app_data_space_id
        if isinstance(customer_data_space_id := item.get("customerDataSpaceId"), str):
            yield SpaceLoader, customer_data_space_id
        for config in cls._get_root_location_configurations(item) or []:
            if isinstance(asset_external_id := config.get("assetExternalId"), str):
                yield AssetLoader, asset_external_id
            if isinstance(data_set_external_id := config.get("dataSetExternalId"), str):
                yield DataSetsLoader, data_set_external_id
            if isinstance(app_data_instance_space := config.get("appDataInstanceSpace"), str):
                yield SpaceLoader, app_data_instance_space
            if isinstance(source_data_instance_space := config.get("sourceDataInstanceSpace"), str):
                yield SpaceLoader, source_data_instance_space
            for key in cls._group_keys:
                for group in config.get(key, []):
                    if isinstance(group, str):
                        yield GroupResourceScopedLoader, group
            data_filters = config.get("dataFilters")
            if not isinstance(data_filters, dict):
                continue
            for key in cls._root_location_filters:
                filter_ = data_filters.get(key)
                if not isinstance(filter_, dict):
                    continue
                for data_set_external_id in filter_.get("dataSetExternalIds", []):
                    if isinstance(data_set_external_id, str):
                        yield DataSetsLoader, data_set_external_id
                for asset_external_id in filter_.get("assetSubtreeExternalIds", []):
                    if isinstance(asset_external_id, str):
                        yield AssetLoader, asset_external_id
                if app_data_instance_space := filter_.get("appDataInstanceSpace"):
                    if isinstance(app_data_instance_space, str):
                        yield SpaceLoader, app_data_instance_space

    def safe_read(self, filepath: Path | str) -> str:
        # The customerDataSpaceVersion is a string, but the user often writes it as an int.
        # YAML will then parse it as an int, for example, `3_0_2` will be parsed as `302`.
        # This is technically a user mistake, as you should quote the version in the YAML file.
        # However, we do not want to put this burden on the user (knowing the intricate workings of YAML),
        # so we fix it here.
        return quote_int_value_by_key_in_yaml(
            safe_read(filepath, encoding=BUILD_FOLDER_ENCODING), key="customerDataSpaceVersion"
        )

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> APMConfigWrite:
        root_location_configurations = self._get_root_location_configurations(resource)
        for config in root_location_configurations or []:
            if not isinstance(config, dict):
                continue
            if ds_external_id := config.pop("dataSetExternalId", None):
                config["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
            data_filters = config.get("dataFilters")
            if not isinstance(data_filters, dict):
                continue
            for key in self._root_location_filters:
                filter_ = data_filters.get(key)
                if not isinstance(filter_, dict):
                    continue
                if ds_external_ids := filter_.pop("dataSetExternalIds", None):
                    filter_["dataSetIds"] = self.client.lookup.data_sets.id(ds_external_ids, is_dry_run)
        return APMConfigWrite._load(resource)

    def dump_resource(self, resource: APMConfig, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        if "existingVersion" not in local:
            # Existing version is typically not set when creating nodes, but we get it back
            # when we retrieve the node from the server.
            dumped.pop("existingVersion", None)

        for config in self._get_root_location_configurations(dumped) or []:
            if not isinstance(config, dict):
                continue
            if data_set_id := config.pop("dataSetId", None):
                config["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
            data_filters = config.get("dataFilters")
            if not isinstance(data_filters, dict):
                continue
            for key in self._root_location_filters:
                filter_ = data_filters.get(key)
                if not isinstance(filter_, dict):
                    continue
                if data_set_ids := filter_.pop("dataSetIds", None):
                    filter_["dataSetExternalIds"] = self.client.lookup.data_sets.external_id(data_set_ids)
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("featureConfiguration", "rootLocationConfigurations"):
            return diff_list_identifiable(local, cdf, get_identifier=hash_dict)
        if not (len(json_path) >= 3 and json_path[:2] == ("featureConfiguration", "rootLocationConfigurations")):
            return super().diff_list(local, cdf, json_path)

        if len(json_path) == 4 and json_path[-1] in self._group_keys:
            return diff_list_hashable(local, cdf)
        if len(json_path) == 5 and json_path[-1] in ("fullWeightModels", "lightWeightModels"):
            return diff_list_identifiable(local, cdf, get_identifier=hash_dict)
        if len(json_path) == 6 and json_path[-2] == "dataFilters" and json_path[-1] in self._root_location_filters:
            return diff_list_hashable(local, cdf)
        if len(json_path) == 7 and "observations" in json_path and json_path[-1] in ("type", "priority"):
            return diff_list_identifiable(local, cdf, get_identifier=hash_dict)
        return super().diff_list(local, cdf, json_path)

    @staticmethod
    def _get_root_location_configurations(resource: dict[str, Any]) -> list | None:
        feature_configuration = resource.get("featureConfiguration")
        if not isinstance(feature_configuration, dict):
            return None
        return feature_configuration.get("rootLocationConfigurations")



import pytest

from cognite_toolkit._cdf_tk.resource_classes import SpaceYAML
from tests.test_unit.utils import find_resources


class TestSpaceYAML:
    @pytest.mark.parametrize("data", list(find_resources("Space")))
    def test_load_valid_space(self, data: dict[str, object]) -> None:
        loaded = SpaceYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data



import contextlib
from collections.abc import Iterator


@contextlib.contextmanager
def tmp_disable_gzip() -> Iterator[None]:
    from cognite.client.config import global_config

    _current_value = global_config.disable_gzip

    global_config.disable_gzip = True
    yield None
    global_config.disable_gzip = _current_value



from __future__ import annotations

import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, ClassVar

from cognite_toolkit._cdf_tk.exceptions import ToolkitFileExistsError

if sys.version_info >= (3, 11):
    import toml
else:
    import tomli as toml


@dataclass(frozen=True)
class ExampleData:
    repo_type: str
    repo: str
    source: str
    destination: Path

    @classmethod
    def load(cls, data: dict[str, Any]) -> ExampleData:
        return cls(
            repo_type=data["repoType"],
            repo=data["repo"],
            source=data["source"],
            destination=data["destination"],
        )


@dataclass(frozen=True)
class ModuleToml:
    filename: ClassVar[str] = "module.toml"
    title: str | None
    tags: frozenset[str] = field(default_factory=frozenset)
    dependencies: frozenset[str] = field(default_factory=frozenset)
    is_selected_by_default: bool = False
    data: list[ExampleData] = field(default_factory=list)
    extra_resources: list[Path] = field(default_factory=list)

    def __post_init__(self) -> None:
        for extra in self.extra_resources:
            if extra.is_absolute():
                raise ToolkitFileExistsError(f"Extra resource {extra} must be a relative path")

    @classmethod
    def load(cls, data: dict[str, Any] | Path) -> ModuleToml:
        if isinstance(data, Path):
            return cls.load(toml.loads(data.read_text(encoding="utf-8")))

        if "dependencies" in data:
            dependencies = frozenset(data["dependencies"].get("modules", set()))
        else:
            dependencies = frozenset()

        example_data: list[ExampleData] = []
        if "data" in data and isinstance(data["data"], list):
            example_data = [ExampleData.load(d) for d in data["data"]]

        extra_resources: list[Path] = []
        if "extra_resources" in data and isinstance(data["extra_resources"], list):
            extra_resources = [Path(item["location"]) for item in data["extra_resources"] if "location" in item]

        tags: frozenset[str] = frozenset()
        if "packages" in data:
            if "tags" in data["packages"]:
                tags = frozenset(data["packages"]["tags"])
        title: str | None = None
        is_selected_by_default: bool = False
        if "module" in data:
            title = data["module"].get("title")
            is_selected_by_default = data["module"].get("is_selected_by_default", False)

        return cls(
            title=title,
            tags=tags,
            dependencies=dependencies,
            is_selected_by_default=is_selected_by_default,
            data=example_data,
            extra_resources=extra_resources,
        )



from ._build_files import BuildDestinationFile, BuildSourceFile
from ._build_variables import BuildVariable, BuildVariables
from ._built_modules import (
    BuiltModule,
    BuiltModuleList,
)
from ._built_resources import (
    BuiltFullResourceList,
    BuiltResource,
    BuiltResourceFull,
    BuiltResourceList,
    SourceLocation,
    SourceLocationEager,
    SourceLocationLazy,
)
from ._config_yaml import (
    BuildConfigYAML,
    BuildEnvironment,
    ConfigEntry,
    ConfigYAMLs,
    Environment,
    InitConfigYAML,
)
from ._deploy_results import (
    DatapointDeployResult,
    DeployResult,
    DeployResults,
    ResourceContainerDeployResult,
    ResourceDeployResult,
    UploadDeployResult,
)
from ._module_directories import ModuleDirectories, ModuleLocation
from ._module_resources import ModuleResources
from ._packages import Package, Packages
from ._yaml_comments import YAMLComments

__all__ = [
    "BuildConfigYAML",
    "BuildDestinationFile",
    "BuildEnvironment",
    "BuildSourceFile",
    "BuildVariable",
    "BuildVariables",
    "BuiltFullResourceList",
    "BuiltModule",
    "BuiltModuleList",
    "BuiltResource",
    "BuiltResourceFull",
    "BuiltResourceList",
    "ConfigEntry",
    "ConfigYAMLs",
    "DatapointDeployResult",
    "DeployResult",
    "DeployResults",
    "Environment",
    "InitConfigYAML",
    "ModuleDirectories",
    "ModuleLocation",
    "ModuleResources",
    "Package",
    "Packages",
    "ResourceContainerDeployResult",
    "ResourceDeployResult",
    "SourceLocation",
    "SourceLocationEager",
    "SourceLocationLazy",
    "UploadDeployResult",
    "YAMLComments",
]



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Any, Optional

from cognite.client import CogniteClient
from cognite.client.data_classes._base import (
    CogniteResourceList,
    InternalIdTransformerMixin,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)

from cognite_toolkit._cdf_tk.client.data_classes.agent_tools import AgentTool


@dataclass
class AgentCore(WriteableCogniteResource["AgentWrite"], ABC):
    """Representation of an AI Agent in CDF.

    Args:
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
        name (str): The name of the agent.
        description (str | None): The description of the agent.
        instructions (str | None): Instructions for the agent.
        model (str | None): Name of the language model to use.
        tools (list[AgentTool] | None): List of tools for the agent.
        labels (list[str] | None): List of labels for the agent.

    """

    external_id: str
    name: str
    description: Optional[str] = None
    instructions: Optional[str] = None
    model: Optional[str] = None
    tools: Optional[list[AgentTool]] = None
    labels: Optional[list[str]] = None

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        result = super().dump(camel_case=camel_case)
        if self.tools:
            result["tools"] = [item.dump(camel_case=camel_case) for item in self.tools]
        if not self.instructions:
            result["instructions"] = ""  # match API behavior
        return result

    def as_write(self) -> AgentWrite:
        return AgentWrite(
            external_id=self.external_id,
            name=self.name,
            description=self.description,
            instructions=self.instructions,
            model=self.model,
            tools=self.tools,
            labels=self.labels,
        )


@dataclass
class Agent(AgentCore):
    """Representation of an AI Agent in CDF.
    This is the read format of an agent.

    Args:
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
        name (str): The name of the agent.
        description (str | None): The description of the agent.
        instructions (str | None): Instructions for the agent.
        model (str | None): Name of the language model to use.
        tools (list[AgentTool] | None): List of tools for the agent.
        labels (list[str] | None): List of labels for the agent.
    """

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: Optional[CogniteClient] = None) -> Agent:
        tools = (
            [AgentTool._load(item) for item in resource.get("tools", [])]
            if isinstance(resource.get("tools"), list)
            else None
        )

        return cls(
            external_id=resource["externalId"],
            name=resource["name"],
            description=resource.get("description"),
            instructions=resource.get("instructions"),
            model=resource.get("model"),
            tools=tools,
            labels=resource.get("labels"),
        )


@dataclass
class AgentWrite(AgentCore):
    """Representation of an AI Agent in CDF.
    This is the write format of an agent.

    Args:
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
        name (str): The name of the agent.
        description (str | None): The description of the agent.
        instructions (str | None): Instructions for the agent.
        model (str | None): Name of the language model to use.
        tools (list[AgentTool] | None): List of tools for the agent.
        labels (list[str] | None): List of labels for the agent.
    """

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: Optional[CogniteClient] = None) -> AgentWrite:
        tools = (
            [AgentTool._load(item) for item in resource.get("tools", [])]
            if isinstance(resource.get("tools"), list)
            else None
        )

        return cls(
            external_id=resource["externalId"],
            name=resource["name"],
            description=resource.get("description", ""),
            instructions=resource.get("instructions", ""),
            model=resource.get("model"),
            tools=tools,
            labels=resource.get("labels", []),
        )


class AgentWriteList(CogniteResourceList[AgentWrite]):
    _RESOURCE = AgentWrite


class AgentList(
    WriteableCogniteResourceList[AgentWrite, Agent],
    InternalIdTransformerMixin,
):
    _RESOURCE = Agent

    def as_write(self) -> AgentWriteList:
        """Returns this AgentList as writeableinstance"""
        return AgentWriteList([item.as_write() for item in self.data], cognite_client=self._get_cognite_client())



import sys
from dataclasses import dataclass
from typing import Any

from cognite.client import CogniteClient
from cognite.client.data_classes._base import CogniteObject
from cognite.client.data_classes.data_modeling import NodeId
from cognite.client.utils._auxiliary import exactly_one_is_not_none

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


@dataclass(frozen=True)
class PendingInstanceId(CogniteObject):
    """A pending instance ID is a NodeId that is set on a TimeSeries or FileMetadata that tells
    the respective syncer that when that Node is created, it should be linked to the
    TimeSeries or FileMetadata.
    """

    pending_instance_id: NodeId
    id: int | None = None
    external_id: str | None = None

    def __post_init__(self) -> None:
        if not exactly_one_is_not_none(self.id, self.external_id):
            raise ValueError("Either id or external_id must be set for PendingInstanceId.")

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        """Load a PendingInstanceId from a dictionary."""
        return cls(
            pending_instance_id=NodeId.load(resource["pendingInstanceId"]),
            id=resource.get("id"),
            external_id=resource.get("externalId"),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Dump the PendingInstanceId to a dictionary."""
        output: dict[str, Any] = {
            "pendingInstanceId" if camel_case else "pending_instance_id": self.pending_instance_id.dump(
                camel_case=camel_case, include_instance_type=False
            ),
        }
        if self.id is not None:
            output["id"] = self.id
        if self.external_id is not None:
            output["externalId"] = self.external_id

        return output



from cognite_toolkit._cdf_tk.client.data_classes.agents import Agent, AgentTool
from tests.test_unit.utils import FakeCogniteResourceGenerator


class TestAgentDataClass:
    def test_agent_data_class_dump_load(self):
        expected = FakeCogniteResourceGenerator(seed=1337).create_instance(Agent)
        dumped = expected.dump()
        actual = Agent._load(dumped)
        assert actual == expected

    def test_agent_tool_data_class_dump_load(self):
        agent_tool = FakeCogniteResourceGenerator(seed=1333).create_instance(AgentTool)
        dumped = agent_tool.dump()
        assert agent_tool == AgentTool._load(dumped)



import re
from abc import abstractmethod
from collections.abc import MutableSequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Literal

from cognite.client.data_classes.data_modeling import DataModelId, ViewId

if TYPE_CHECKING:
    pass


class GraphQLParser:
    _token_pattern = re.compile(r"[\w\n]+|[^\w\s]", flags=re.DOTALL)
    _multi_newline = re.compile(r"\n+")

    def __init__(self, raw: str, data_model_id: DataModelId) -> None:
        # Ensure consistent line endings
        self.raw = raw.replace("\r\n", "\n").replace("\r", "\n")
        self.data_model_id = data_model_id
        self._entities: list[_Entity] | None = None

    def get_views(self, include_version: bool = False) -> set[ViewId]:
        views: set[ViewId] = set()
        for entity in self._get_entities():
            if entity.is_imported:
                continue
            view_directive: _ViewDirective | None = None
            for directive in entity.directives:
                if isinstance(directive, _ViewDirective):
                    view_directive = directive
                    break
            if view_directive:
                views.add(
                    ViewId(
                        view_directive.space or self.data_model_id.space,
                        view_directive.external_id or entity.identifier,
                        version=view_directive.version if include_version else None,
                    )
                )
            else:
                views.add(ViewId(self.data_model_id.space, entity.identifier))
        return views

    def get_dependencies(self, include_version: bool = False) -> set[ViewId | DataModelId]:
        dependencies: set[ViewId | DataModelId] = set()
        for entity in self._get_entities():
            view_directive: _ViewDirective | None = None
            is_dependency = False
            for directive in entity.directives:
                if isinstance(directive, _Import):
                    if directive.data_model:
                        dependencies.add(directive.data_model)
                        break
                    is_dependency = True
                elif isinstance(directive, _ViewDirective):
                    view_directive = directive
            if is_dependency and view_directive:
                dependencies.add(
                    ViewId(
                        view_directive.space or self.data_model_id.space,
                        view_directive.external_id or entity.identifier,
                        version=view_directive.version if include_version else None,
                    )
                )
            elif is_dependency:
                # Todo: Warning Likely invalid directive
                ...
        return dependencies

    def _get_entities(self) -> "list[_Entity]":
        if self._entities is None:
            self._entities = self._parse()
        return self._entities

    def _parse(self) -> "list[_Entity]":
        entities: list[_Entity] = []
        entity: _Entity | None = None
        last_class: Literal["type", "interface"] | None = None

        parentheses: list[str] = []
        directive_tokens: _DirectiveTokens | None = None
        is_directive_start = False
        is_multiline_comment = False
        is_end_of_line_comment = False
        is_in_double_quote = False
        is_in_single_quote = False
        tokens = self._token_pattern.findall(self.raw)
        for no, token in enumerate(tokens):
            if no >= 2 and (tokens[no - 2 : no + 1] == ['"'] * 3 or tokens[no - 2 : no + 1] == ["'"] * 3):
                is_multiline_comment = not is_multiline_comment
            elif token == '"':
                is_in_double_quote = not is_in_double_quote
            elif token == "'":
                is_in_single_quote = not is_in_single_quote
            elif token == "#" and not (is_in_double_quote or is_in_single_quote):
                is_end_of_line_comment = True
            if "\n" in token and is_end_of_line_comment:
                is_end_of_line_comment = False
            if is_multiline_comment or is_end_of_line_comment:
                continue

            token = self._multi_newline.sub("\n", token)
            if token != "\n":
                token = token.removesuffix("\n").removeprefix("\n")

            if token in "({[<":
                parentheses.append(token)
            elif token in ")}]>":
                parentheses.pop()

            is_end_of_entity = bool(parentheses and parentheses[0] == "{")
            if entity and is_end_of_entity:
                # End of entity definition
                if directive_tokens and (directive := directive_tokens.create()):
                    entity.directives.append(directive)
                directive_tokens = None
                entities.append(entity)
                entity = None
            elif entity is not None:
                if directive_tokens is not None and (not parentheses or token == "@"):
                    # End of directive
                    if directive := directive_tokens.create():
                        entity.directives.append(directive)
                    directive_tokens = None
                    if token == "@":
                        is_directive_start = True
                elif directive_tokens:
                    # Gather the content of the directive
                    directive_tokens.append(token)
                elif token == "@":
                    is_directive_start = True
                elif is_directive_start:
                    if token in ("import", "view"):
                        # We only care about import and view directives
                        directive_tokens = _DirectiveTokens([token])
                    is_directive_start = False

            elif token in ("type", "interface") and not parentheses:
                # Next token starts a new entity definition
                # Notet hat we cannot be inside a paranthesis as that could be a
                # property of the entity
                last_class = token
            elif last_class is not None and token != "\n":
                # Start of a new entity definition
                entity = _Entity(identifier=token, class_=last_class)
                last_class = None
        return entities


class _DirectiveTokens(list, MutableSequence[str]):
    def create(self) -> "_Directive | None":
        return _Directive.load(self)


@dataclass
class _Directive:
    # This pattern ignores commas inside }
    SPLIT_ON_COMMA_PATTERN = re.compile(r",(?![^{]*\})")

    @classmethod
    def load(cls, content: list[str]) -> "_Directive | None":
        key, *content = content
        raw_string = cls._standardize(content)
        data = cls._create_args(raw_string)
        if isinstance(data, list):
            return None
        if key == "import":
            return _Import._load(data)
        if key == "view":
            return _ViewDirective._load(data)
        return None

    @classmethod
    def _standardize(cls, content: list[str]) -> str:
        """We standardize to use commas as separators, instead of newlines.
        However, if we are inside a parenthesis we need to replace newlines with empty.
        """
        if not content:
            return ""
        # Ensure that the content is wrapped in parenthesis
        # so that we can safely drop the first and last character
        if content[0] != "(":
            content.insert(0, "(")
        if content[-1] != ")":
            content.append(")")

        standardized: list[str] = []
        for last, current, next_ in zip(content, content[1:], content[2:]):
            if current == "\n" and last in ")}]" and next_ in "({[":
                standardized.append(",")
            elif current == "\n" and last in ":({[":
                continue
            elif current == "\n" and next_ in ")}]":
                continue
            elif current == "\n":
                standardized.append(",")
            else:
                standardized.append(current)
        return "".join(standardized)

    @classmethod
    def _create_args(cls, string: str) -> dict[str, Any] | str | list[Any]:
        if "," not in string and ":" not in string:
            return string
        if string[0] == "{" and string[-1] == "}":
            string = string[1:-1]
        is_list = False
        if string[0] == "[" and string[-1] == "]":
            string = string[1:-1]
            is_list = True
        items: list[Any] = []
        obj: dict[str, Any] = {}
        last_pair = ""
        for pair in cls.SPLIT_ON_COMMA_PATTERN.split(string):
            stripped = pair.strip()
            if (not stripped or (not is_list and ":" not in stripped)) and not last_pair:
                continue
            if last_pair:
                stripped = f"{last_pair},{stripped}"
                last_pair = ""
            # Regex does not deal with nested parenthesis
            left_count = sum(stripped.count(char) for char in "{[(")
            right_count = sum(stripped.count(char) for char in "}])")
            if left_count != right_count:
                last_pair = stripped
                continue
            if is_list:
                items.append(cls._create_args(stripped))
            else:
                key, value = cls._clean(*stripped.split(":", maxsplit=1))
                if set("{[(}]}") & set(key):
                    raise ValueError(f"Invalid value {value}")
                obj[key] = cls._create_args(value)
        return items if is_list else obj

    @classmethod
    def _clean(cls, *args: Any) -> Any:
        return tuple(arg.removeprefix('"').removesuffix('"').removeprefix('"').removesuffix('"') for arg in args)

    @classmethod
    @abstractmethod
    def _load(cls, data: dict[str, Any]) -> "_Directive": ...


@dataclass
class _ViewDirective(_Directive):
    space: str | None = None
    external_id: str | None = None
    version: str | None = None

    @classmethod
    def _load(cls, data: dict[str, Any] | str) -> "_ViewDirective":
        if isinstance(data, str):
            return _ViewDirective()
        space = data.get("space")
        external_id = data.get("externalId")
        version = data.get("version")
        for variable in (space, external_id, version):
            if variable and not isinstance(variable, str):
                raise ValueError(f"Invalid variable {variable}")
        return _ViewDirective(space=space, external_id=external_id, version=version)


@dataclass
class _Import(_Directive):
    data_model: DataModelId | None = None

    @classmethod
    def _load(cls, data: dict[str, Any] | str) -> "_Import":
        if isinstance(data, str):
            return _Import()
        if "dataModel" in data:
            return _Import(data_model=DataModelId.load(data["dataModel"]))
        return _Import()


@dataclass
class _Entity:
    identifier: str
    class_: Literal["type", "interface"]
    directives: list[_Directive] = field(default_factory=list)

    @property
    def is_imported(self) -> bool:
        return any(isinstance(directive, _Import) for directive in self.directives)



"""Toolkit resource classes are pydantic models that represent the YAML file format the Cognite Toolkit uses.

This is means that we have three set of resource classes we use in Toolkit:

1. Toolkit resource classes (this module): Represent the YAML configuration format the Toolkit uses.
2. Write resource classes (from cognite-sdk): Represent the write/request format the Cognite resources.
3. Read resource classes (from cognite-sdk): Represent the read/response format the Cognite resources.
"""

from .asset import AssetYAML
from .base import BaseModelResource, ToolkitResource
from .dataset import DataSetYAML
from .event import EventYAML
from .filemetadata import FileMetadataYAML
from .function_schedule import FunctionScheduleYAML
from .functions import FunctionsYAML
from .groups import GroupYAML
from .labels import LabelsYAML
from .location import LocationYAML
from .raw_database_table import DatabaseYAML, TableYAML
from .securitycategories import SecurityCategoriesYAML
from .space import SpaceYAML
from .threedmodels import ThreeDModelYAML
from .timeseries import TimeSeriesYAML
from .transformation_schedule import TransformationScheduleYAML
from .transformations import TransformationYAML

__all__ = [
    "AssetYAML",
    "BaseModelResource",
    "DataSetYAML",
    "DatabaseYAML",
    "EventYAML",
    "FileMetadataYAML",
    "FunctionScheduleYAML",
    "FunctionsYAML",
    "GroupYAML",
    "LabelsYAML",
    "LocationYAML",
    "SecurityCategoriesYAML",
    "SpaceYAML",
    "TableYAML",
    "ThreeDModelYAML",
    "TimeSeriesYAML",
    "ToolkitResource",
    "TransformationScheduleYAML",
    "TransformationYAML",
]



from typing import Literal

from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError

from ._base import ToolkitCommand


class CollectCommand(ToolkitCommand):
    def execute(self, action: Literal["opt-in", "opt-out"]) -> None:
        if action == "opt-in":
            self.tracker.enable()
            print("You have successfully opted in to data collection.")
        elif action == "opt-out":
            self.tracker.disable()
            print("You have successfully opted out of data collection.")
        else:
            raise ToolkitValueError(f"Invalid action: {action}. Must be 'opt-in' or 'opt-out'.")



from __future__ import annotations

from collections.abc import Hashable, Iterable, Sequence
from functools import lru_cache
from pathlib import Path
from typing import Any, final

from cognite.client.data_classes.capabilities import Capability, LocationFiltersAcl
from cognite.client.data_classes.data_modeling import DataModelId, ViewId
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk._parameters import ANY_INT, ParameterSpec, ParameterSpecSet
from cognite_toolkit._cdf_tk.client.data_classes.location_filters import (
    LocationFilter,
    LocationFilterList,
    LocationFilterWrite,
    LocationFilterWriteList,
)
from cognite_toolkit._cdf_tk.constants import BUILD_FOLDER_ENCODING
from cognite_toolkit._cdf_tk.exceptions import ResourceRetrievalError
from cognite_toolkit._cdf_tk.loaders._base_loaders import ResourceLoader
from cognite_toolkit._cdf_tk.utils import in_dict, quote_int_value_by_key_in_yaml, safe_read
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_hashable, diff_list_identifiable, dm_identifier

from .classic_loaders import AssetLoader, SequenceLoader
from .data_organization_loaders import DataSetsLoader
from .datamodel_loaders import DataModelLoader, SpaceLoader, ViewLoader
from .file_loader import FileMetadataLoader
from .timeseries_loaders import TimeSeriesLoader


@final
class LocationFilterLoader(
    ResourceLoader[str, LocationFilterWrite, LocationFilter, LocationFilterWriteList, LocationFilterList]
):
    folder_name = "locations"
    filename_pattern = r"^.*LocationFilter$"
    resource_cls = LocationFilter
    resource_write_cls = LocationFilterWrite
    list_cls = LocationFilterList
    list_write_cls = LocationFilterWriteList
    dependencies = frozenset(
        {
            AssetLoader,
            DataSetsLoader,
            DataModelLoader,
            SpaceLoader,
            ViewLoader,
            SequenceLoader,
            FileMetadataLoader,
            TimeSeriesLoader,
        }
    )
    kind = "LocationFilter"
    _doc_base_url = "https://api-docs.cogheim.net/redoc/#tag/"
    _doc_url = "Location-Filters/operation/createLocationFilter"

    subfilter_names = ("assets", "events", "files", "timeseries", "sequences")

    @property
    def display_name(self) -> str:
        return "location filters"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[LocationFilterWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        # Todo: Specify space ID scopes:

        actions = (
            [LocationFiltersAcl.Action.Read]
            if read_only
            else [LocationFiltersAcl.Action.Read, LocationFiltersAcl.Action.Write]
        )

        return LocationFiltersAcl(
            actions=actions,
            scope=LocationFiltersAcl.Scope.All(),
            allow_unknown=True,
        )

    @classmethod
    def get_id(cls, item: LocationFilter | LocationFilterWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if not item.external_id:
            raise KeyError("LocationFilter must have external_id")
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    def safe_read(self, filepath: Path | str) -> str:
        # The version is a string, but the user often writes it as an int.
        # YAML will then parse it as an int, for example, `3_0_2` will be parsed as `302`.
        # This is technically a user mistake, as you should quote the version in the YAML file.
        # However, we do not want to put this burden on the user (knowing the intricate workings of YAML),
        # so we fix it here.
        return quote_int_value_by_key_in_yaml(safe_read(filepath, encoding=BUILD_FOLDER_ENCODING), key="version")

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> LocationFilterWrite:
        if parent_external_id := resource.pop("parentExternalId", None):
            # This is a workaround: when the parentExternalId cannot be resolved because the parent hasn't been created yet, , we save it so that we can try again "later"

            #
            try:
                resource["parentId"] = self.client.lookup.location_filters.id(parent_external_id, is_dry_run)
            except ResourceRetrievalError:
                resource["parentId"] = -1
                resource["_parentExternalId"] = parent_external_id

        if "assetCentric" not in resource:
            return LocationFilterWrite._load(resource)
        asset_centric = resource["assetCentric"]
        if data_set_external_ids := asset_centric.pop("dataSetExternalIds", None):
            asset_centric["dataSetIds"] = self.client.lookup.data_sets.id(
                data_set_external_ids, is_dry_run, allow_empty=True
            )
        for subfilter_name in self.subfilter_names:
            subfilter = asset_centric.get(subfilter_name, {})
            if data_set_external_ids := subfilter.pop("dataSetExternalIds", []):
                asset_centric[subfilter_name]["dataSetIds"] = self.client.lookup.data_sets.id(
                    data_set_external_ids,
                    is_dry_run,
                    allow_empty=True,
                )

        return LocationFilterWrite._load(resource)

    def dump_resource(self, resource: LocationFilter, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        if parent_id := dumped.pop("parentId", None):
            dumped["parentExternalId"] = self.client.lookup.location_filters.external_id(parent_id)
        if "dataModelingType" in dumped and "dataModelingType" not in local:
            # Default set on server side
            dumped.pop("dataModelingType")
        if "assetCentric" not in dumped:
            return dumped
        asset_centric = dumped["assetCentric"]
        if data_set_ids := asset_centric.pop("dataSetIds", None):
            asset_centric["dataSetExternalIds"] = self.client.lookup.data_sets.external_id(data_set_ids)
        for subfilter_name in self.subfilter_names:
            subfilter = asset_centric.get(subfilter_name, {})
            if data_set_ids := subfilter.pop("dataSetIds", []):
                asset_centric[subfilter_name]["dataSetExternalIds"] = self.client.lookup.data_sets.external_id(
                    data_set_ids
                )
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path[0] == "assetCentric" or json_path == ("instanceSpaces",):
            return diff_list_hashable(local, cdf)
        elif json_path in [("dataModels",), ("views",), ("scene",)]:
            return diff_list_identifiable(local, cdf, get_identifier=dm_identifier)
        return super().diff_list(local, cdf, json_path)

    def create(self, items: LocationFilterWrite | LocationFilterWriteList) -> LocationFilterList:
        if isinstance(items, LocationFilterWrite):
            items = LocationFilterWriteList([items])

        created = []
        # Note: the Location API does not support batch creation, so we need to do this one by one.
        # Furthermore, we could not do the parentExternalId->parentId lookup before the parent was created,
        # hence it may be deferred here.
        for item in items:
            # These are set if lookup has been deferred
            if item._parent_external_id and item.parent_id == -1:
                item.parent_id = self.client.lookup.location_filters.id(item._parent_external_id)
            created.append(self.client.location_filters.create(item))
        return LocationFilterList(created)

    def retrieve(self, external_ids: SequenceNotStr[str]) -> LocationFilterList:
        all_locations = self.client.location_filters.list()
        found_locations: LocationFilterList = LocationFilterList([])

        # locationfilter list returns a tree structure, so we need to traverse it
        def _recursive_find(locs: LocationFilterList) -> None:
            for loc in locs:
                if loc.external_id in external_ids:
                    found_locations.append(loc)
                if loc.locations:
                    _recursive_find(loc.locations)

        _recursive_find(all_locations)
        return LocationFilterList(found_locations)

    def update(self, items: LocationFilterWrite | LocationFilterWriteList) -> LocationFilterList:
        if isinstance(items, LocationFilterWrite):
            items = LocationFilterWriteList([items])

        updated = []
        ids = {item.external_id: item.id for item in self.retrieve([item.external_id for item in items])}
        for update in items:
            updated.append(self.client.location_filters.update(ids[update.external_id], update))
        return LocationFilterList(updated)

    def delete(self, external_ids: SequenceNotStr[str]) -> int:
        count = 0
        for id in [loc.id for loc in self.retrieve(external_ids)]:
            self.client.location_filters.delete(id)
            count += 1
        return count

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[LocationFilter]:
        return iter(self.client.location_filters)

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        spec = super().get_write_cls_parameter_spec()
        # Added by toolkit
        spec.add(
            ParameterSpec(
                ("parentExternalId",),
                frozenset({"str"}),
                is_required=False,
                _is_nullable=True,
            )
        )
        spec.discard(
            ParameterSpec(
                ("parentId",),
                frozenset({"int"}),
                is_required=False,
                _is_nullable=True,
            )
        )
        spec.add(
            ParameterSpec(
                (
                    "assetCentric",
                    "dataSetExternalIds",
                ),
                frozenset({"list"}),
                is_required=False,
                _is_nullable=False,
            )
        )
        spec.add(
            ParameterSpec(
                (
                    "assetCentric",
                    "dataSetExternalIds",
                    ANY_INT,
                ),
                frozenset({"str"}),
                is_required=False,
                _is_nullable=False,
            )
        )
        spec.discard(
            ParameterSpec(
                (
                    "assetCentric",
                    "dataSetIds",
                ),
                frozenset({"list"}),
                is_required=False,
                _is_nullable=False,
            )
        )
        spec.discard(
            ParameterSpec(
                (
                    "assetCentric",
                    "dataSetIds",
                    ANY_INT,
                ),
                frozenset({"int"}),
                is_required=False,
                _is_nullable=False,
            )
        )
        spec.discard(
            ParameterSpec(
                ("assetCentric", "assetSubtreeIds", ANY_INT, "externalId"), frozenset({"str", "int"}), False, False
            )
        )
        spec.add(
            ParameterSpec(("assetCentric", "assetSubtreeIds", ANY_INT, "externalId"), frozenset({"str"}), False, False)
        )
        spec.add(ParameterSpec(("assetCentric", "assetSubtreeIds", ANY_INT, "id"), frozenset({"int"}), False, False))
        for subfilter_name in cls.subfilter_names:
            spec.discard(
                ParameterSpec(
                    ("assetCentric", subfilter_name, "assetSubtreeIds", ANY_INT, "externalId"),
                    frozenset({"str", "int"}),
                    False,
                    False,
                )
            )
            spec.add(
                ParameterSpec(
                    ("assetCentric", subfilter_name, "assetSubtreeIds", ANY_INT, "externalId"),
                    frozenset({"str"}),
                    False,
                    False,
                )
            )
            spec.add(
                ParameterSpec(
                    ("assetCentric", subfilter_name, "assetSubtreeIds", ANY_INT, "id"), frozenset({"int"}), False, False
                )
            )
            spec.add(
                ParameterSpec(("assetCentric", subfilter_name, "dataSetExternalIds"), frozenset({"list"}), False, False)
            )
            spec.add(
                ParameterSpec(
                    ("assetCentric", subfilter_name, "dataSetExternalIds", ANY_INT), frozenset({"str"}), False, False
                )
            )
            spec.discard(
                ParameterSpec(("assetCentric", subfilter_name, "dataSetIds"), frozenset({"list"}), False, False)
            )
            spec.discard(
                ParameterSpec(("assetCentric", subfilter_name, "dataSetIds", ANY_INT), frozenset({"int"}), False, False)
            )

        spec.add(ParameterSpec(("dataModels", ANY_INT, "type"), frozenset({"str"}), False, False))
        spec.add(ParameterSpec(("views", ANY_INT, "type"), frozenset({"str"}), False, False))
        return spec

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        """Returns all items that this item requires.

        For example, a TimeSeries requires a DataSet, so this method would return the
        DatasetLoader and identifier of that dataset.
        """
        if "assetCentric" in item:
            asset_centric = item["assetCentric"]
            for data_set_external_id in asset_centric.get("dataSetExternalIds", []):
                yield DataSetsLoader, data_set_external_id
            for asset in asset_centric.get("assetSubtreeIds", []):
                if "externalId" in asset:
                    yield AssetLoader, asset["externalId"]
            for subfilter_name in cls.subfilter_names:
                subfilter = asset_centric.get(subfilter_name, {})
                for data_set_external_id in subfilter.get("dataSetExternalIds", []):
                    yield DataSetsLoader, data_set_external_id
                for asset in subfilter.get("assetSubtreeIds", []):
                    if "externalId" in asset:
                        yield AssetLoader, asset["externalId"]
        for view in item.get("views", []):
            if in_dict(["space", "externalId", "version"], view):
                yield ViewLoader, ViewId(view["space"], view["externalId"], view["version"])
        for space in item.get("instanceSpaces", []):
            yield SpaceLoader, space
        for data_model in item.get("dataModels", []):
            if in_dict(["space", "externalId", "version"], data_model):
                yield DataModelLoader, DataModelId(data_model["space"], data_model["externalId"], data_model["version"])



from __future__ import annotations

import enum
import inspect
import itertools
import types
import typing
from collections.abc import Iterable, Mapping, MutableMapping, MutableSequence, Sequence
from typing import Any, get_origin

from .constants import BASE_TYPES, TYPES


class TypeHint:
    _DICT_TYPES = {dict, typing.Dict, MutableSequence, MutableMapping, typing.Mapping, Mapping}  # noqa UP006
    _LIST_TYPES = {list, typing.Sequence, Sequence, typing.List, MutableSequence}  # noqa UP006

    def __init__(self, raw: Any) -> None:
        self.raw = raw
        self._container_type = get_origin(raw)
        self.is_union = self._is_union(self._container_type)
        if self.is_union:
            self.args = typing.get_args(raw)
        else:
            self.args = (raw,)

    def __str__(self) -> str:
        return ",".join(self.types)

    @classmethod
    def _is_union(cls, x: Any) -> bool:
        try:
            return x in [types.UnionType, typing.Union]  # type: ignore[attr-defined]
        except AttributeError:
            # Python 3.9
            return x is typing.Union

    @classmethod
    def _is_none_type(cls, x: Any) -> bool:
        try:
            return x in [None, types.NoneType]  # type: ignore[attr-defined]
        except AttributeError:
            # Python 3.9
            return x is None

    @classmethod
    def _as_str(cls, arg: Any) -> str:
        try:
            value = arg.__name__
        except AttributeError:
            # Python 3.9
            value = str(arg).removeprefix("typing.")
            if "[" in value:
                value = value.split("[")[0]
        if value.casefold() in TYPES:
            return value.casefold()
        elif value == "Literal":
            return "str"
        elif value == "Sequence":
            return "list"
        elif value == "Any":
            return "unknown"
        elif inspect.isclass(arg) and issubclass(arg, enum.Enum):
            return "str"
        return "dict"

    @property
    def types(self) -> list[str]:
        return list({self._as_str(arg) for arg in self.args if not self._is_none_type(arg)})

    @property
    def frozen_types(self) -> frozenset[str]:
        return frozenset(self.types)

    @property
    def is_base_type(self) -> bool:
        return any(type_ in BASE_TYPES for type_ in self.types)

    @property
    def is_nullable(self) -> bool:
        return any(self._is_none_type(arg) for arg in self.args)

    @property
    def is_any(self) -> bool:
        return any(arg is typing.Any for arg in self.args)

    @property
    def is_user_defined_class(self) -> bool:
        return any(
            inspect.isclass(arg) and arg.__module__ not in {"typing", "builtins", "collections.abc"}
            for arg in self.args
        )

    @property
    def _get_origins(self) -> tuple[Any, ...]:
        return tuple(get_origin(arg) for arg in self.args)

    @property
    def is_dict_type(self) -> bool:
        return any(arg in self._DICT_TYPES for arg in itertools.chain(self._get_origins, self.args))

    @property
    def is_list_type(self) -> bool:
        return any(arg in self._LIST_TYPES for arg in itertools.chain(self._get_origins, self.args))

    @property
    def container_args(self) -> tuple[Any, ...]:
        return typing.get_args(self.args[0])

    def __repr__(self) -> str:
        return repr(self.raw)

    @property
    def sub_hints(self) -> Iterable[TypeHint]:
        for arg in self.args:
            if self._is_none_type(arg):
                continue
            yield TypeHint(arg)



from pathlib import Path
from unittest.mock import MagicMock

import pytest

from cognite_toolkit._cdf_tk.builders._location import LocationBuilder
from cognite_toolkit._cdf_tk.data_classes._build_files import BuildSourceFile
from cognite_toolkit._cdf_tk.data_classes._module_directories import ModuleLocation
from cognite_toolkit._cdf_tk.exceptions import ToolkitError


def test_location_builder_detect_cyclic_references(build_tmp_path):
    # Setup

    module_location = MagicMock(spec=ModuleLocation)

    source_file = MagicMock(spec=BuildSourceFile)
    source_file.source = MagicMock()
    source_file.source.path = Path("cyclic.LocationFilter.yaml")
    source_file.loaded = [
        {"externalId": "location1", "description": "Location 1", "parentExternalId": "location2"},
        {"externalId": "location2", "description": "Location 2", "parentExternalId": "location3"},
        {
            "externalId": "location3",
            "description": "Location 3",
            "parentExternalId": "location1",  # This creates a cycle
        },
    ]

    location_builder = LocationBuilder(build_dir=build_tmp_path)
    with pytest.raises(ToolkitError, match="Circular dependency found in Locations*"):
        list(location_builder.build(source_files=[source_file], module=module_location))


def test_location_builder_detect_self_reference(build_tmp_path):
    # Setup

    module_location = MagicMock(spec=ModuleLocation)

    source_file = MagicMock(spec=BuildSourceFile)
    source_file.source = MagicMock()
    source_file.source.path = Path("self.LocationFilter.yaml")
    source_file.loaded = [
        {
            "externalId": "location1",
            "description": "Location 1",
            "parentExternalId": "location1",
        },  # This creates a self-reference
    ]

    location_builder = LocationBuilder(build_dir=build_tmp_path)
    with pytest.raises(ToolkitError, match="Circular dependency found in Locations*"):
        list(location_builder.build(source_files=[source_file], module=module_location))



import pytest
from cognite.client.data_classes.data_modeling import Node

from cognite_toolkit._cdf_tk.client.data_classes.apm_config_v1 import APMConfig, APMConfigCore, APMConfigWrite
from tests.test_unit.utils import FakeCogniteResourceGenerator


class TestAPMConfigV1Class:
    def test_apm_config_data_class_dump_load(self) -> None:
        config = FakeCogniteResourceGenerator(seed=1337).create_instance(APMConfig)

        dumped = config.dump()
        assert config == APMConfig._load(dumped)

    def test_apm_config_write_data_class_dump_load(self) -> None:
        config = FakeCogniteResourceGenerator(seed=1333).create_instance(APMConfigWrite)

        dumped = config.dump()
        assert config == APMConfigWrite._load(dumped)

    def test_load_unknown_field_in_feature_configuration(self) -> None:
        config = FakeCogniteResourceGenerator(seed=1337).create_instance(APMConfig)
        dumped = config.dump()
        unknown_object = {
            "with_unknown_field": True,
            "unknown_field": "unknown_value",
        }
        dumped["featureConfiguration"]["unknown_field"] = unknown_object

        loaded = APMConfig._load(dumped)
        redumped = loaded.dump()
        assert redumped["featureConfiguration"]["unknown_field"] == unknown_object

    def test_load_unknown_nested_field_in_feature_configuration(self) -> None:
        config = FakeCogniteResourceGenerator(seed=1337).create_instance(APMConfig)
        dumped = config.dump()
        unknown_object = {
            "with_unknown_field": True,
            "unknown_field": "unknown_value",
        }
        assert isinstance(dumped["featureConfiguration"]["rootLocationConfigurations"], list)
        dumped["featureConfiguration"]["rootLocationConfigurations"].append(unknown_object)

        loaded = APMConfig._load(dumped)
        redumped = loaded.dump()
        assert redumped["featureConfiguration"]["rootLocationConfigurations"][-1] == unknown_object

    def test_from_node_as_write_to_node(self) -> None:
        node = Node._load(
            {
                "space": "APM_Config",
                "externalId": "my_config",
                "version": 1,
                "lastUpdatedTime": 1,
                "createdTime": 1,
                "properties": {
                    "APM_Config": {
                        "APM_Config/1": {
                            "customerDataSpaceId": "APM_SourceData",
                            "customerDataSpaceVersion": "1",
                            "name": "Infield APM App Config",
                            "featureConfiguration": {
                                "rootLocationConfigurations": [
                                    {
                                        "assetExternalId": "MyAsset",
                                        "appDataInstanceSpace": "my_instance_space",
                                        "sourceDataInstanceSpace": "my_source_space",
                                        "templateAdmins": ["admin1", "admin2"],
                                        "checklistAdmins": ["admin3"],
                                    },
                                ],
                            },
                        }
                    }
                },
            }
        )

        read_config = APMConfig.from_node(node)

        write_config = read_config.as_write()

        node_apply = write_config.as_node()

        assert node_apply.dump() == node.as_write().dump()

    @pytest.mark.parametrize("config_cls", [APMConfig, APMConfigWrite])
    def test_apm_config_write_dump_load_yaml(self, config_cls: type[APMConfigCore]) -> None:
        instance = FakeCogniteResourceGenerator(seed=1338).create_instance(config_cls)

        yaml_str = instance.dump_yaml()

        loaded_instance = config_cls.load(yaml_str)

        assert loaded_instance.dump() == instance.dump()



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Any, Optional

from cognite.client import CogniteClient
from cognite.client.data_classes._base import (
    CogniteResourceList,
    InternalIdTransformerMixin,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)


@dataclass
class AgentToolCore(WriteableCogniteResource["AgentToolWrite"], ABC):
    """
    Representation of an AI Agent Tool in CDF.

    Args:
        name (str): The name of the tool.
        type (str): The type of the tool.
        description (str): The description of the tool.
        configuration (dict[str, Any] | None): The configuration of the tool.
    """

    name: str
    type: str
    description: str
    configuration: Optional[dict[str, Any]] = None

    def as_write(self) -> AgentToolWrite:
        return AgentToolWrite(
            name=self.name,
            type=self.type,
            description=self.description,
            configuration=self.configuration,
        )


@dataclass
class AgentTool(AgentToolCore):
    """
    Representation of an AI Agent Tool in CDF.
    This is the read format of an agent tool.

    Args:
        name (str): The name of the tool.
        type (str): The type of the tool.
        description (str): The description of the tool.
        configuration (dict[str, Any] | None): The configuration of the tool.
    """

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: Optional[CogniteClient] = None) -> AgentTool:
        return cls(
            name=resource["name"],
            type=resource["type"],
            description=resource["description"],
            configuration=resource.get("configuration"),
        )


@dataclass
class AgentToolWrite(AgentToolCore):
    """Representation of an AI Agent Tool in CDF.
    This is the write format of an agent tool.

    Args:
        name (str): The name of the tool.
        type (str): The type of the tool.
        description (str): The description of the tool.
        configuration (dict[str, Any] | None): The configuration of the tool.
    """

    pass


class AgentToolWriteList(CogniteResourceList[AgentToolWrite]):
    _RESOURCE = AgentToolWrite


class AgentToolList(
    WriteableCogniteResourceList[AgentToolWrite, AgentTool],
    InternalIdTransformerMixin,
):
    _RESOURCE = AgentTool

    def as_write(self) -> AgentToolWriteList:
        """Returns this AgentToolWriteList instance"""
        return AgentToolWriteList([item.as_write() for item in self.data], cognite_client=self._get_cognite_client())



from __future__ import annotations

from cognite.client.data_classes import (
    Asset,
    AssetList,
    AssetWrite,
    AssetWriteList,
    Database,
    DatabaseList,
    Datapoints,
    DatapointsList,
    DatapointSubscription,
    DatapointSubscriptionList,
    DataPointSubscriptionWrite,
    DatapointSubscriptionWriteList,
    DataSet,
    DataSetList,
    DataSetWrite,
    DataSetWriteList,
    Event,
    EventList,
    EventWrite,
    EventWriteList,
    ExtractionPipeline,
    ExtractionPipelineConfig,
    ExtractionPipelineConfigWrite,
    ExtractionPipelineConfigWriteList,
    ExtractionPipelineList,
    ExtractionPipelineWrite,
    ExtractionPipelineWriteList,
    FileMetadata,
    FileMetadataList,
    FileMetadataWrite,
    FileMetadataWriteList,
    Function,
    FunctionList,
    FunctionSchedule,
    FunctionSchedulesList,
    FunctionScheduleWrite,
    FunctionScheduleWriteList,
    FunctionWrite,
    FunctionWriteList,
    Group,
    GroupList,
    GroupWrite,
    GroupWriteList,
    LabelDefinition,
    LabelDefinitionList,
    LabelDefinitionWrite,
    Relationship,
    RelationshipList,
    RelationshipWrite,
    RelationshipWriteList,
    Row,
    RowList,
    RowWrite,
    RowWriteList,
    SecurityCategory,
    SecurityCategoryList,
    SecurityCategoryWrite,
    SecurityCategoryWriteList,
    Sequence,
    SequenceList,
    SequenceRows,
    SequenceRowsList,
    SequenceWrite,
    SequenceWriteList,
    Table,
    TableList,
    TableWrite,
    TableWriteList,
    ThreeDModel,
    ThreeDModelList,
    ThreeDModelWrite,
    ThreeDModelWriteList,
    TimeSeries,
    TimeSeriesList,
    TimeSeriesWrite,
    TimeSeriesWriteList,
    Transformation,
    TransformationList,
    TransformationNotification,
    TransformationNotificationList,
    TransformationSchedule,
    TransformationScheduleList,
    TransformationScheduleWrite,
    TransformationScheduleWriteList,
    TransformationWrite,
    TransformationWriteList,
    Workflow,
    WorkflowList,
    WorkflowTrigger,
    WorkflowTriggerList,
    WorkflowTriggerUpsert,
    WorkflowTriggerUpsertList,
    WorkflowUpsert,
    WorkflowUpsertList,
    WorkflowVersion,
    WorkflowVersionList,
    WorkflowVersionUpsert,
    WorkflowVersionUpsertList,
)
from cognite.client.data_classes.data_modeling import (
    Container,
    ContainerApply,
    ContainerApplyList,
    ContainerList,
    DataModel,
    DataModelApply,
    DataModelApplyList,
    DataModelList,
    Node,
    NodeApply,
    NodeApplyList,
    NodeList,
    Space,
    SpaceApply,
    SpaceApplyList,
    SpaceList,
    View,
    ViewApply,
    ViewApplyList,
    ViewList,
)
from cognite.client.data_classes.extractionpipelines import ExtractionPipelineConfigList
from cognite.client.data_classes.hosted_extractors import (
    Destination,
    DestinationList,
    DestinationWrite,
    DestinationWriteList,
    Job,
    JobList,
    JobWrite,
    JobWriteList,
    Mapping,
    MappingList,
    MappingWrite,
    MappingWriteList,
    Source,
    SourceList,
    SourceWrite,
    SourceWriteList,
)
from cognite.client.data_classes.iam import TokenInspection
from cognite.client.data_classes.labels import LabelDefinitionWriteList
from cognite.client.data_classes.transformations.notifications import (
    TransformationNotificationWrite,
    TransformationNotificationWriteList,
)

from cognite_toolkit._cdf_tk.client.data_classes import (
    robotics,
)
from cognite_toolkit._cdf_tk.client.data_classes.graphql_data_models import (
    GraphQLDataModel,
    GraphQLDataModelList,
    GraphQLDataModelWrite,
    GraphQLDataModelWriteList,
)
from cognite_toolkit._cdf_tk.client.data_classes.location_filters import (
    LocationFilter,
    LocationFilterList,
    LocationFilterWrite,
    LocationFilterWriteList,
)
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawDatabase, RawDatabaseList

from .data_classes import APIResource, Method

# This is used to define the resources that should be mocked in the ApprovalCogniteClient
# You can add more resources here if you need to mock more resources
API_RESOURCES = [
    APIResource(
        api_name="post",
        resource_cls=TokenInspection,
        list_cls=list[TokenInspection],
        methods={
            "post": [Method(api_class_method="post", mock_class_method="post_method")],
        },
    ),
    APIResource(
        api_name="iam.groups",
        resource_cls=Group,
        _write_cls=GroupWrite,
        _write_list_cls=GroupWriteList,
        list_cls=GroupList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "delete": [Method(api_class_method="delete", mock_class_method="delete_id_external_id")],
            "retrieve": [Method(api_class_method="list", mock_class_method="return_values")],
        },
    ),
    APIResource(
        api_name="iam.token",
        resource_cls=TokenInspection,
        list_cls=list[TokenInspection],
        methods={
            "inspect": [Method(api_class_method="inspect", mock_class_method="return_value")],
        },
    ),
    APIResource(
        api_name="data_sets",
        resource_cls=DataSet,
        _write_cls=DataSetWrite,
        _write_list_cls=DataSetWriteList,
        list_cls=DataSetList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "retrieve": [
                Method(api_class_method="list", mock_class_method="return_values"),
                Method(api_class_method="retrieve", mock_class_method="return_value"),
                Method(api_class_method="retrieve_multiple", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="time_series",
        resource_cls=TimeSeries,
        _write_cls=TimeSeriesWrite,
        list_cls=TimeSeriesList,
        _write_list_cls=TimeSeriesWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "delete": [Method(api_class_method="delete", mock_class_method="delete_id_external_id")],
            "retrieve": [
                Method(api_class_method="list", mock_class_method="return_values"),
                Method(api_class_method="retrieve", mock_class_method="return_values"),
                Method(api_class_method="retrieve_multiple", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="raw.databases",
        resource_cls=Database,
        _write_cls=RawDatabase,
        list_cls=DatabaseList,
        _write_list_cls=RawDatabaseList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "retrieve": [Method(api_class_method="list", mock_class_method="return_values")],
            "delete": [Method(api_class_method="delete", mock_class_method="delete_raw")],
        },
    ),
    APIResource(
        api_name="raw.tables",
        resource_cls=Table,
        _write_cls=TableWrite,
        list_cls=TableList,
        _write_list_cls=TableWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_raw_table")],
            "retrieve": [Method(api_class_method="list", mock_class_method="return_values")],
            "delete": [Method(api_class_method="delete", mock_class_method="delete_raw")],
        },
    ),
    APIResource(
        api_name="raw.rows",
        resource_cls=Row,
        _write_cls=RowWrite,
        list_cls=RowList,
        _write_list_cls=RowWriteList,
        methods={
            "create": [Method(api_class_method="insert_dataframe", mock_class_method="insert_dataframe")],
            "delete": [Method(api_class_method="delete", mock_class_method="delete_raw")],
            "retrieve": [
                Method(api_class_method="list", mock_class_method="return_values"),
                Method(api_class_method="retrieve", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="functions",
        resource_cls=Function,
        _write_cls=FunctionWrite,
        list_cls=FunctionList,
        _write_list_cls=FunctionWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_single")],
            "delete": [Method(api_class_method="delete", mock_class_method="delete_id_external_id")],
            "retrieve": [
                Method(api_class_method="list", mock_class_method="return_values"),
                Method(api_class_method="retrieve", mock_class_method="return_value"),
                Method(api_class_method="retrieve_multiple", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="functions.schedules",
        resource_cls=FunctionSchedule,
        _write_cls=FunctionScheduleWrite,
        list_cls=FunctionSchedulesList,
        _write_list_cls=FunctionScheduleWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
        },
    ),
    APIResource(
        api_name="transformations",
        resource_cls=Transformation,
        _write_cls=TransformationWrite,
        list_cls=TransformationList,
        _write_list_cls=TransformationWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "delete": [Method(api_class_method="delete", mock_class_method="delete_id_external_id")],
            "retrieve": [
                Method(api_class_method="list", mock_class_method="return_values"),
                Method(api_class_method="retrieve", mock_class_method="return_value"),
                Method(api_class_method="retrieve_multiple", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="transformations.schedules",
        resource_cls=TransformationSchedule,
        _write_cls=TransformationScheduleWrite,
        list_cls=TransformationScheduleList,
        _write_list_cls=TransformationScheduleWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "delete": [Method(api_class_method="delete", mock_class_method="delete_id_external_id")],
            "retrieve": [
                Method(api_class_method="list", mock_class_method="return_values"),
                Method(api_class_method="retrieve", mock_class_method="return_value"),
            ],
        },
    ),
    APIResource(
        api_name="extraction_pipelines",
        resource_cls=ExtractionPipeline,
        _write_cls=ExtractionPipelineWrite,
        list_cls=ExtractionPipelineList,
        _write_list_cls=ExtractionPipelineWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "delete": [Method(api_class_method="delete", mock_class_method="delete_id_external_id")],
            "retrieve": [
                Method(api_class_method="list", mock_class_method="return_values"),
                Method(api_class_method="retrieve", mock_class_method="return_value"),
                Method(api_class_method="retrieve_multiple", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="extraction_pipelines.config",
        resource_cls=ExtractionPipelineConfig,
        _write_cls=ExtractionPipelineConfigWrite,
        list_cls=ExtractionPipelineConfigList,
        _write_list_cls=ExtractionPipelineConfigWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_extraction_pipeline_config")],
            "retrieve": [
                Method(api_class_method="list", mock_class_method="return_values"),
                Method(api_class_method="retrieve", mock_class_method="return_value"),
            ],
        },
    ),
    APIResource(
        api_name="data_modeling.containers",
        resource_cls=Container,
        list_cls=ContainerList,
        _write_cls=ContainerApply,
        _write_list_cls=ContainerApplyList,
        methods={
            "create": [Method(api_class_method="apply", mock_class_method="create_multiple")],
            "delete": [Method(api_class_method="delete", mock_class_method="delete_data_modeling")],
            "retrieve": [
                Method(api_class_method="list", mock_class_method="return_values"),
                Method(api_class_method="retrieve", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="data_modeling.views",
        resource_cls=View,
        list_cls=ViewList,
        _write_cls=ViewApply,
        _write_list_cls=ViewApplyList,
        methods={
            "create": [Method(api_class_method="apply", mock_class_method="create_multiple")],
            "delete": [Method(api_class_method="delete", mock_class_method="delete_data_modeling")],
            "retrieve": [
                Method(api_class_method="list", mock_class_method="return_values"),
                Method(api_class_method="retrieve", mock_class_method="data_model_retrieve"),
            ],
        },
    ),
    APIResource(
        api_name="data_modeling.data_models",
        resource_cls=DataModel,
        list_cls=DataModelList,
        _write_cls=DataModelApply,
        _write_list_cls=DataModelApplyList,
        methods={
            "create": [Method(api_class_method="apply", mock_class_method="create_multiple")],
            "delete": [Method(api_class_method="delete", mock_class_method="delete_data_modeling")],
            "retrieve": [
                Method(api_class_method="list", mock_class_method="return_values"),
                Method(api_class_method="retrieve", mock_class_method="return_data_models"),
            ],
        },
    ),
    APIResource(
        api_name="data_modeling.spaces",
        resource_cls=Space,
        list_cls=SpaceList,
        _write_cls=SpaceApply,
        _write_list_cls=SpaceApplyList,
        methods={
            "create": [Method(api_class_method="apply", mock_class_method="create_multiple")],
            "delete": [Method(api_class_method="delete", mock_class_method="delete_space")],
            "retrieve": [
                Method(api_class_method="list", mock_class_method="return_values"),
                Method(api_class_method="retrieve", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="time_series.data",
        resource_cls=Datapoints,
        list_cls=DatapointsList,
        methods={
            "create": [
                Method(api_class_method="insert", mock_class_method="create_multiple"),
                Method(api_class_method="insert_dataframe", mock_class_method="insert_dataframe"),
            ],
        },
    ),
    APIResource(
        api_name="files",
        resource_cls=FileMetadata,
        list_cls=FileMetadataList,
        _write_cls=FileMetadataWrite,
        _write_list_cls=FileMetadataWriteList,
        methods={
            "create": [
                Method(api_class_method="upload", mock_class_method="upload"),
                Method(api_class_method="create", mock_class_method="create_filemetadata"),
                # This is used by functions to upload the file used for deployment.
                Method(api_class_method="upload_bytes", mock_class_method="upload_bytes_files_api"),
                Method(
                    api_class_method="upload_content_bytes", mock_class_method="upload_file_content_bytes_files_api"
                ),
                Method(api_class_method="upload_content", mock_class_method="upload_file_content_path_files_api"),
            ],
            "delete": [Method(api_class_method="delete", mock_class_method="delete_id_external_id")],
            "retrieve": [
                Method(api_class_method="list", mock_class_method="return_values"),
                Method(api_class_method="retrieve", mock_class_method="files_retrieve"),
                Method(api_class_method="retrieve_multiple", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="data_modeling.instances",
        resource_cls=Node,
        list_cls=NodeList,
        _write_cls=NodeApply,
        _write_list_cls=NodeApplyList,
        methods={
            "create": [Method(api_class_method="apply", mock_class_method="create_instances")],
            "delete": [Method(api_class_method="delete", mock_class_method="delete_instances")],
            "retrieve": [
                Method(api_class_method="list", mock_class_method="return_values"),
                Method(api_class_method="retrieve", mock_class_method="return_instances"),
            ],
        },
    ),
    APIResource(
        api_name="workflows",
        resource_cls=Workflow,
        list_cls=WorkflowList,
        _write_cls=WorkflowUpsert,
        _write_list_cls=WorkflowUpsertList,
        methods={
            "create": [Method(api_class_method="upsert", mock_class_method="upsert")],
            # "update": [Method(api_class_method="upsert", mock_name="upsert")],
            # "delete": [Method(api_class_method="delete", mock_name="delete_id_external_id")],
            "retrieve": [
                Method(api_class_method="retrieve", mock_class_method="return_value"),
            ],
        },
    ),
    APIResource(
        api_name="workflows.versions",
        resource_cls=WorkflowVersion,
        list_cls=WorkflowVersionList,
        _write_cls=WorkflowVersionUpsert,
        _write_list_cls=WorkflowVersionUpsertList,
        methods={
            "create": [Method(api_class_method="upsert", mock_class_method="upsert")],
            # "update": [Method(api_class_method="upsert", mock_name="upsert")],
            # "delete": [Method(api_class_method="delete", mock_name="delete")],
            "retrieve": [
                Method(api_class_method="retrieve", mock_class_method="return_value"),
                Method(api_class_method="list", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="iam.security_categories",
        resource_cls=SecurityCategory,
        list_cls=SecurityCategoryList,
        _write_cls=SecurityCategoryWrite,
        _write_list_cls=SecurityCategoryWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "retrieve": [Method(api_class_method="list", mock_class_method="return_values")],
        },
    ),
    APIResource(
        api_name="time_series.subscriptions",
        resource_cls=DatapointSubscription,
        list_cls=DatapointSubscriptionList,
        _write_cls=DataPointSubscriptionWrite,
        _write_list_cls=DatapointSubscriptionWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_single")],
            "retrieve": [
                Method(api_class_method="retrieve", mock_class_method="return_value"),
            ],
        },
    ),
    APIResource(
        api_name="labels",
        resource_cls=LabelDefinition,
        list_cls=LabelDefinitionList,
        _write_cls=LabelDefinitionWrite,
        _write_list_cls=LabelDefinitionWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "retrieve": [Method(api_class_method="list", mock_class_method="return_values")],
        },
    ),
    APIResource(
        api_name="transformations.notifications",
        resource_cls=TransformationNotification,
        list_cls=TransformationNotificationList,
        _write_cls=TransformationNotificationWrite,
        _write_list_cls=TransformationNotificationWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "retrieve": [Method(api_class_method="list", mock_class_method="return_values")],
        },
    ),
    APIResource(
        api_name="assets",
        resource_cls=Asset,
        list_cls=AssetList,
        _write_cls=AssetWrite,
        _write_list_cls=AssetWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "retrieve": [
                Method(api_class_method="list", mock_class_method="return_values"),
                Method(api_class_method="retrieve_multiple", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="three_d.models",
        resource_cls=ThreeDModel,
        list_cls=ThreeDModelList,
        _write_cls=ThreeDModelWrite,
        _write_list_cls=ThreeDModelWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_3dmodel")],
            "retrieve": [
                Method(api_class_method="__iter__", mock_class_method="iterate_values"),
            ],
        },
    ),
    APIResource(
        api_name="robotics.data_postprocessing",
        resource_cls=robotics.DataPostProcessing,
        list_cls=robotics.DataPostProcessingList,
        _write_cls=robotics.DataPostProcessingWrite,
        _write_list_cls=robotics.DataPostProcessingWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "retrieve": [
                Method(api_class_method="retrieve", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="robotics.frames",
        resource_cls=robotics.Frame,
        list_cls=robotics.FrameList,
        _write_cls=robotics.FrameWrite,
        _write_list_cls=robotics.FrameWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "retrieve": [
                Method(api_class_method="retrieve", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="robotics.locations",
        resource_cls=robotics.Location,
        list_cls=robotics.LocationList,
        _write_cls=robotics.LocationWrite,
        _write_list_cls=robotics.LocationWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "retrieve": [
                Method(api_class_method="retrieve", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="robotics.maps",
        resource_cls=robotics.Map,
        list_cls=robotics.MapList,
        _write_cls=robotics.MapWrite,
        _write_list_cls=robotics.MapWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "retrieve": [
                Method(api_class_method="retrieve", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="robotics.capabilities",
        resource_cls=robotics.RobotCapability,
        list_cls=robotics.RobotCapabilityList,
        _write_cls=robotics.RobotCapabilityWrite,
        _write_list_cls=robotics.RobotCapabilityWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "retrieve": [
                Method(api_class_method="retrieve", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="location_filters",
        resource_cls=LocationFilter,
        list_cls=LocationFilterList,
        _write_cls=LocationFilterWrite,
        _write_list_cls=LocationFilterWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_single")],
            "retrieve": [
                Method(api_class_method="retrieve", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="dml",
        resource_cls=GraphQLDataModel,
        list_cls=GraphQLDataModelList,
        _write_cls=GraphQLDataModelWrite,
        _write_list_cls=GraphQLDataModelWriteList,
        methods={
            "create": [Method(api_class_method="apply_dml", mock_class_method="apply_dml")],
        },
    ),
    APIResource(
        api_name="sequences",
        resource_cls=Sequence,
        list_cls=SequenceList,
        _write_cls=SequenceWrite,
        _write_list_cls=SequenceWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "retrieve": [
                Method(api_class_method="retrieve_multiple", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="sequences.rows",
        resource_cls=SequenceRows,
        list_cls=SequenceRowsList,
        _write_cls=SequenceRows,
        _write_list_cls=SequenceRowsList,
        methods={
            "create": [Method(api_class_method="insert", mock_class_method="create_single")],
            "retrieve": [
                Method(api_class_method="retrieve", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="workflows.triggers",
        resource_cls=WorkflowTrigger,
        list_cls=WorkflowTriggerList,
        _write_cls=WorkflowTriggerUpsert,
        _write_list_cls=WorkflowTriggerUpsertList,
        methods={
            "create": [Method(api_class_method="upsert", mock_class_method="create_multiple")],
            "retrieve": [
                Method(api_class_method="list", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="hosted_extractors.sources",
        resource_cls=Source,
        list_cls=SourceList,
        _write_cls=SourceWrite,
        _write_list_cls=SourceWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "retrieve": [
                Method(api_class_method="retrieve", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="hosted_extractors.destinations",
        resource_cls=Destination,
        list_cls=DestinationList,
        _write_cls=DestinationWrite,
        _write_list_cls=DestinationWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "retrieve": [
                Method(api_class_method="retrieve", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="hosted_extractors.jobs",
        resource_cls=Job,
        list_cls=JobList,
        _write_cls=JobWrite,
        _write_list_cls=JobWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "retrieve": [
                Method(api_class_method="retrieve", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="hosted_extractors.mappings",
        resource_cls=Mapping,
        list_cls=MappingList,
        _write_cls=MappingWrite,
        _write_list_cls=MappingWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "retrieve": [
                Method(api_class_method="retrieve", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="events",
        resource_cls=Event,
        list_cls=EventList,
        _write_cls=EventWrite,
        _write_list_cls=EventWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "retrieve": [
                Method(api_class_method="retrieve_multiple", mock_class_method="return_values"),
            ],
        },
    ),
    APIResource(
        api_name="relationships",
        resource_cls=Relationship,
        list_cls=RelationshipList,
        _write_cls=RelationshipWrite,
        _write_list_cls=RelationshipWriteList,
        methods={
            "create": [Method(api_class_method="create", mock_class_method="create_multiple")],
            "retrieve": [
                Method(api_class_method="retrieve_multiple", mock_class_method="return_values"),
            ],
        },
    ),
]



from collections.abc import Hashable
from pathlib import Path
from unittest.mock import MagicMock

import pytest
from cognite.client.data_classes import data_modeling as dm

from cognite_toolkit._cdf_tk._parameters import read_parameters_from_dict
from cognite_toolkit._cdf_tk.loaders import ContainerLoader, ResourceLoader, ResourceWorker, SpaceLoader, ViewLoader
from tests.test_unit.approval_client import ApprovalToolkitClient


class TestViewLoader:
    @pytest.mark.parametrize(
        "item",
        [
            pytest.param(
                {
                    "filter": {
                        "hasData": [
                            {"type": "container", "space": "sp_my_space", "externalId": "container_id"},
                            {"type": "view", "space": "sp_my_space", "externalId": "view_id"},
                        ]
                    }
                },
                id="HasData Filter",
            ),
            pytest.param(
                {
                    "properties": {
                        "reverseDirectRelation": {
                            "connectionType": "multi_reverse_direct_relation",
                            "source": {
                                "type": "view",
                                "space": "sp_my_space",
                                "externalId": "view_id",
                                "version": "v42",
                            },
                            "through": {
                                "source": {
                                    "type": "view",
                                    "space": "sp_my_space",
                                    "externalId": "view_id",
                                    "version": "v42",
                                },
                                "identifier": "view_property",
                            },
                        }
                    }
                },
                id="Reverse Direct Relation Property",
            ),
        ],
    )
    def test_valid_spec(self, item: dict):
        spec = ViewLoader.get_write_cls_parameter_spec()
        dumped = read_parameters_from_dict(item)

        extra = dumped - spec

        assert not extra, f"Extra keys: {extra}"

    def test_unchanged_view_int_version(self, toolkit_client_approval: ApprovalToolkitClient) -> None:
        loader = ViewLoader.create_loader(toolkit_client_approval.mock_client)
        raw_file = """- space: sp_space
  externalId: my_view
  version: 1"""
        file = MagicMock(spec=Path)
        file.read_text.return_value = raw_file
        cdf_view = dm.View(
            space="sp_space",
            external_id="my_view",
            version="1",
            last_updated_time=1,
            created_time=1,
            description=None,
            name=None,
            filter=None,
            implements=None,
            writable=True,
            used_for="node",
            is_global=False,
            properties={},
        )

        toolkit_client_approval.append(dm.View, [cdf_view])

        worker = ResourceWorker(loader)
        to_create, to_change, to_delete, unchanged, _ = worker.load_resources([file])
        assert {
            "create": len(to_create),
            "change": len(to_change),
            "delete": len(to_delete),
            "unchanged": len(unchanged),
        } == {"create": 0, "change": 0, "delete": 0, "unchanged": 1}

    @pytest.mark.parametrize(
        "item, expected",
        [
            pytest.param(
                {
                    "space": "sp_my_space",
                    "properties": {
                        "name": {
                            "container": {
                                "type": "container",
                                "space": "my_container_space",
                                "externalId": "my_container",
                            }
                        }
                    },
                },
                [
                    (SpaceLoader, "sp_my_space"),
                    (ContainerLoader, dm.ContainerId(space="my_container_space", external_id="my_container")),
                ],
                id="View with one container property",
            ),
            pytest.param(
                {
                    "space": "sp_my_space",
                    "properties": {
                        "toEdge": {
                            "source": {
                                "type": "view",
                                "space": "my_view_space",
                                "externalId": "my_view",
                                "version": "1",
                            },
                            "edgeSource": {
                                "type": "view",
                                "space": "my_other_view_space",
                                "externalId": "my_edge_view",
                                "version": "42",
                            },
                        }
                    },
                },
                [
                    (SpaceLoader, "sp_my_space"),
                    (ViewLoader, dm.ViewId(space="my_view_space", external_id="my_view", version="1")),
                    (ViewLoader, dm.ViewId(space="my_other_view_space", external_id="my_edge_view", version="42")),
                ],
                id="View with one container property",
            ),
        ],
    )
    def test_get_dependent_items(self, item: dict, expected: list[tuple[type[ResourceLoader], Hashable]]) -> None:
        actual = ViewLoader.get_dependent_items(item)

        assert list(actual) == expected



from __future__ import annotations

from collections.abc import Hashable, Iterable, Sequence
from functools import lru_cache
from pathlib import Path
from typing import Any

from cognite.client.data_classes import ClientCredentials
from cognite.client.data_classes.capabilities import Capability, HostedExtractorsAcl
from cognite.client.data_classes.hosted_extractors import (
    Destination,
    DestinationList,
    DestinationWrite,
    DestinationWriteList,
    EventHubSourceWrite,
    Job,
    JobList,
    JobWrite,
    JobWriteList,
    KafkaSourceWrite,
    Mapping,
    MappingList,
    MappingWrite,
    MappingWriteList,
    Source,
    SourceList,
    SourceWrite,
    SourceWriteList,
)
from cognite.client.data_classes.hosted_extractors.sources import (
    AuthenticationWrite,
    BasicAuthenticationWrite,
    RESTClientCredentialsAuthenticationWrite,
    RestSourceWrite,
    _MQTTSourceWrite,
)
from cognite.client.utils.useful_types import SequenceNotStr
from rich.console import Console

from cognite_toolkit._cdf_tk._parameters import ANYTHING, ParameterSpec, ParameterSpecSet
from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.exceptions import ToolkitNotSupported
from cognite_toolkit._cdf_tk.loaders._base_loaders import ResourceLoader
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning

from .data_organization_loaders import DataSetsLoader


class HostedExtractorSourceLoader(ResourceLoader[str, SourceWrite, Source, SourceWriteList, SourceList]):
    folder_name = "hosted_extractors"
    filename_pattern = r".*\.Source$"  # Matches all yaml files whose stem ends with '.Source'.
    resource_cls = Source
    resource_write_cls = SourceWrite
    list_cls = SourceList
    list_write_cls = SourceWriteList
    kind = "Source"
    _doc_base_url = "https://api-docs.cognite.com/20230101-alpha/tag/"
    _doc_url = "Sources/operation/create_sources"
    _SupportedSources = (_MQTTSourceWrite, KafkaSourceWrite, RestSourceWrite, EventHubSourceWrite)

    @property
    def display_name(self) -> str:
        return "hosted extractor sources"

    @classmethod
    def get_id(cls, item: SourceWrite | Source | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: Sequence[SourceWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [HostedExtractorsAcl.Action.Read]
            if read_only
            else [HostedExtractorsAcl.Action.Read, HostedExtractorsAcl.Action.Write]
        )

        return HostedExtractorsAcl(
            actions,
            HostedExtractorsAcl.Scope.All(),
        )

    def create(self, items: SourceWriteList) -> SourceList:
        return self.client.hosted_extractors.sources.create(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> SourceList:
        return self.client.hosted_extractors.sources.retrieve(external_ids=ids, ignore_unknown_ids=True)

    def update(self, items: SourceWriteList) -> SourceList:
        return self.client.hosted_extractors.sources.update(items, mode="replace")

    def delete(self, ids: SequenceNotStr[str]) -> int:
        self.client.hosted_extractors.sources.delete(ids, ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Source]:
        return iter(self.client.hosted_extractors.sources)

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        # parameterspec is highly dependent on type of source, so we accept any parameter
        return ParameterSpecSet(
            [ParameterSpec((ANYTHING,), frozenset({"dict"}), is_required=False, _is_nullable=False)]
        )

    def dump_resource(self, resource: Source, local: dict[str, Any] | None = None) -> dict[str, Any]:
        HighSeverityWarning(
            "Sources will always be considered different, and thus will always be redeployed."
        ).print_warning()
        return self.dump_id(resource.external_id)

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> SourceWrite:
        loaded = super().load_resource(resource, is_dry_run=is_dry_run)
        if not isinstance(loaded, self._SupportedSources):
            # We need to explicitly check for the supported types as we need to ensure that we know
            # what the sensitive strings are for each type of source.
            # Technically, we could support any new source type added to the cognite-sdk, but
            # we would risk printing sensitive strings in the terminal.
            raise ToolkitNotSupported(
                f"Hosted extractor source type {type(loaded).__name__} is not supported."
                f"Please contact support to request support for this source type."
            )
        return loaded

    def sensitive_strings(self, item: SourceWrite) -> Iterable[str]:
        if isinstance(item, _MQTTSourceWrite | KafkaSourceWrite | RestSourceWrite) and item.authentication:
            yield from self._sensitive_auth_strings(item.authentication)
        if (
            isinstance(item, _MQTTSourceWrite | KafkaSourceWrite)
            and item.auth_certificate
            and item.auth_certificate.key_password
        ):
            yield item.auth_certificate.key_password

    @staticmethod
    def _sensitive_auth_strings(auth: AuthenticationWrite) -> Iterable[str]:
        if isinstance(auth, BasicAuthenticationWrite):
            yield auth.password
        elif isinstance(auth, RESTClientCredentialsAuthenticationWrite):
            yield auth.client_secret


class HostedExtractorDestinationLoader(
    ResourceLoader[str, DestinationWrite, Destination, DestinationWriteList, DestinationList]
):
    folder_name = "hosted_extractors"
    filename_pattern = r".*\.Destination$"  # Matches all yaml files whose stem ends with '.Destination'.
    resource_cls = Destination
    resource_write_cls = DestinationWrite
    list_cls = DestinationList
    list_write_cls = DestinationWriteList
    dependencies = frozenset({DataSetsLoader})
    kind = "Destination"
    _doc_base_url = "https://api-docs.cognite.com/20230101-alpha/tag/"
    _doc_url = "Destinations/operation/create_destinations"

    def __init__(self, client: ToolkitClient, build_dir: Path | None, console: Console | None = None):
        super().__init__(client, build_dir, console)
        self._authentication_by_id: dict[str, ClientCredentials] = {}

    @property
    def display_name(self) -> str:
        return "hosted extractor destinations"

    @classmethod
    def get_id(cls, item: DestinationWrite | Destination | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: Sequence[DestinationWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [HostedExtractorsAcl.Action.Read]
            if read_only
            else [HostedExtractorsAcl.Action.Read, HostedExtractorsAcl.Action.Write]
        )

        return HostedExtractorsAcl(
            actions,
            HostedExtractorsAcl.Scope.All(),
        )

    def create(self, items: DestinationWriteList) -> DestinationList:
        return self.client.hosted_extractors.destinations.create(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> DestinationList:
        return self.client.hosted_extractors.destinations.retrieve(external_ids=ids, ignore_unknown_ids=True)

    def update(self, items: DestinationWriteList) -> DestinationList:
        return self.client.hosted_extractors.destinations.update(items, mode="replace")

    def delete(self, ids: SequenceNotStr[str]) -> int:
        self.client.hosted_extractors.destinations.delete(ids, ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Destination]:
        return iter(self.client.hosted_extractors.destinations)

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> DestinationWrite:
        if raw_auth := resource.pop("credentials", None):
            credentials = ClientCredentials._load(raw_auth)
            self._authentication_by_id[self.get_id(resource)] = credentials
            if is_dry_run:
                resource["credentials"] = {"nonce": "dummy_nonce"}
            else:
                session = self.client.iam.sessions.create(credentials, "CLIENT_CREDENTIALS")
                resource["credentials"] = {"nonce": session.nonce}
        if ds_external_id := resource.pop("targetDataSetExternalId", None):
            resource["targetDataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        return DestinationWrite._load(resource)

    def dump_resource(self, resource: Destination, local: dict[str, Any] | None = None) -> dict[str, Any]:
        HighSeverityWarning(
            "Destinations will always be considered different, and thus will always be redeployed."
        ).print_warning()
        return self.dump_id(resource.external_id)

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        spec = super().get_write_cls_parameter_spec()
        # Handled by Toolkit

        spec.add(ParameterSpec(("credentials", "clientId"), frozenset({"str"}), is_required=False, _is_nullable=True))
        spec.add(
            ParameterSpec(("credentials", "clientSecret"), frozenset({"str"}), is_required=False, _is_nullable=True)
        )

        spec.discard(ParameterSpec(("targetDataSetId",), frozenset({"int"}), is_required=False, _is_nullable=True))
        spec.add(ParameterSpec(("targetDataSetExternalId",), frozenset({"str"}), is_required=False, _is_nullable=True))
        return spec

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        if "targetDataSetId" in item:
            yield DataSetsLoader, item["targetDataSetId"]

    def sensitive_strings(self, item: DestinationWrite) -> Iterable[str]:
        yield item.credentials.nonce
        id_ = self.get_id(item)
        if id_ in self._authentication_by_id:
            yield self._authentication_by_id[id_].client_secret


class HostedExtractorJobLoader(ResourceLoader[str, JobWrite, Job, JobWriteList, JobList]):
    folder_name = "hosted_extractors"
    filename_pattern = r".*\.Job$"  # Matches all yaml files whose stem ends with '.Job'.
    resource_cls = Job
    resource_write_cls = JobWrite
    list_cls = JobList
    list_write_cls = JobWriteList
    dependencies = frozenset({HostedExtractorSourceLoader, HostedExtractorDestinationLoader})
    kind = "Job"
    _doc_base_url = "https://api-docs.cognite.com/20230101-alpha/tag/"
    _doc_url = "Jobs/operation/create_jobs"

    @property
    def display_name(self) -> str:
        return "hosted extractor jobs"

    @classmethod
    def get_id(cls, item: JobWrite | Job | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: Sequence[JobWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [HostedExtractorsAcl.Action.Read]
            if read_only
            else [HostedExtractorsAcl.Action.Read, HostedExtractorsAcl.Action.Write]
        )

        return HostedExtractorsAcl(
            actions,
            HostedExtractorsAcl.Scope.All(),
        )

    def dump_resource(self, resource: Job, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        if not dumped.get("config") and "config" not in local:
            dumped.pop("config", None)
        return dumped

    def create(self, items: JobWriteList) -> JobList:
        return self.client.hosted_extractors.jobs.create(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> JobList:
        return self.client.hosted_extractors.jobs.retrieve(external_ids=ids, ignore_unknown_ids=True)

    def update(self, items: JobWriteList) -> JobList:
        return self.client.hosted_extractors.jobs.update(items, mode="replace")

    def delete(self, ids: SequenceNotStr[str]) -> int:
        self.client.hosted_extractors.jobs.delete(ids, ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Job]:
        return iter(self.client.hosted_extractors.jobs)

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        spec = super().get_write_cls_parameter_spec()
        # Used by the SDK to determine the class to load
        spec.add(
            ParameterSpec(
                (
                    "format",
                    "type",
                ),
                frozenset({"str"}),
                is_required=True,
                _is_nullable=False,
            )
        )
        spec.add(
            ParameterSpec(
                ("config", "incrementalLoad", "type"), frozenset({"str"}), is_required=True, _is_nullable=False
            )
        )
        spec.add(
            ParameterSpec(("config", "pagination", "type"), frozenset({"str"}), is_required=True, _is_nullable=False)
        )

        return spec

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceLoader], Hashable]]:
        if "sourceId" in item:
            yield HostedExtractorSourceLoader, item["sourceId"]
        if "destinationId" in item:
            yield HostedExtractorDestinationLoader, item["destinationId"]


class HostedExtractorMappingLoader(ResourceLoader[str, MappingWrite, Mapping, MappingWriteList, MappingList]):
    folder_name = "hosted_extractors"
    filename_pattern = r".*\.Mapping$"  # Matches all yaml files whose stem ends with '.Mapping'.
    resource_cls = Mapping
    resource_write_cls = MappingWrite
    list_cls = MappingList
    list_write_cls = MappingWriteList
    # This is not an explicit dependency, however, adding it here as mapping will should be deployed after source.
    dependencies = frozenset({HostedExtractorSourceLoader})
    kind = "Mapping"
    _doc_base_url = "https://api-docs.cognite.com/20230101-alpha/tag/"
    _doc_url = "Mappings/operation/create_mappings"

    @property
    def display_name(self) -> str:
        return "hosted extractor mappings"

    @classmethod
    def get_id(cls, item: MappingWrite | Mapping | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: Sequence[MappingWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [HostedExtractorsAcl.Action.Read]
            if read_only
            else [HostedExtractorsAcl.Action.Read, HostedExtractorsAcl.Action.Write]
        )

        return HostedExtractorsAcl(
            actions,
            HostedExtractorsAcl.Scope.All(),
        )

    def create(self, items: MappingWriteList) -> MappingList:
        return self.client.hosted_extractors.mappings.create(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> MappingList:
        return self.client.hosted_extractors.mappings.retrieve(external_ids=ids, ignore_unknown_ids=True)

    def update(self, items: MappingWriteList) -> MappingList:
        return self.client.hosted_extractors.mappings.update(items)

    def delete(self, ids: SequenceNotStr[str]) -> int:
        self.client.hosted_extractors.mappings.delete(ids, ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Mapping]:
        return iter(self.client.hosted_extractors.mappings)

    @classmethod
    @lru_cache(maxsize=1)
    def get_write_cls_parameter_spec(cls) -> ParameterSpecSet:
        spec = super().get_write_cls_parameter_spec()
        # Used by the SDK to determine the class to load
        spec.add(ParameterSpec(("input", "type"), frozenset({"str"}), is_required=True, _is_nullable=False))
        return spec



import itertools
from collections.abc import Iterable, Iterator
from pathlib import Path
from typing import Literal, overload

from cognite_toolkit._cdf_tk.constants import (
    BUILTIN_MODULES,
    MODULE_PATH_SEP,
    ROOT_MODULES,
)


def iterate_modules(root_dir: Path) -> Iterator[tuple[Path, list[Path]]]:
    """Iterate over all modules in the project and yield the module directory and all files in the module.

    Args:
        root_dir (Path): The root directory of the project

    Yields:
        Iterator[tuple[Path, list[Path]]]: A tuple containing the module directory and a list of all files in the module

    """
    if root_dir.name in ROOT_MODULES:
        yield from _iterate_modules(root_dir)
        return
    elif root_dir.name == BUILTIN_MODULES:
        yield from _iterate_modules(root_dir)
        return
    for root_module in ROOT_MODULES:
        module_dir = root_dir / root_module
        if module_dir.exists():
            yield from _iterate_modules(module_dir)


def _iterate_modules(root_dir: Path) -> Iterator[tuple[Path, list[Path]]]:
    # local import to avoid circular import
    from cognite_toolkit._cdf_tk.constants import EXCL_FILES
    from cognite_toolkit._cdf_tk.loaders import LOADER_BY_FOLDER_NAME

    if not root_dir.exists():
        return
    for module_dir in root_dir.iterdir():
        if not module_dir.is_dir():
            continue
        sub_directories = [path for path in module_dir.iterdir() if path.is_dir()]
        is_any_resource_directories = any(dir.name in LOADER_BY_FOLDER_NAME for dir in sub_directories)
        if sub_directories and is_any_resource_directories:
            # Module found
            yield module_dir, [path for path in module_dir.rglob("*") if path.is_file() and path.name not in EXCL_FILES]
            # Stop searching for modules in subdirectories
            continue
        yield from _iterate_modules(module_dir)


@overload
def module_from_path(path: Path, return_resource_folder: Literal[True]) -> tuple[str, str]: ...


@overload
def module_from_path(path: Path, return_resource_folder: Literal[False] = False) -> str: ...


def module_from_path(path: Path, return_resource_folder: bool = False) -> str | tuple[str, str]:
    """Get the module name from a path"""
    # local import to avoid circular import
    from cognite_toolkit._cdf_tk.loaders import LOADER_BY_FOLDER_NAME

    if len(path.parts) == 1:
        raise ValueError("Path is not a module")
    last_folder = path.parts[1]
    for part in path.parts[1:]:
        if part in LOADER_BY_FOLDER_NAME:
            if return_resource_folder:
                return last_folder, part
            return last_folder
        last_folder = part
    raise ValueError("Path is not part of a module")


def module_directory_from_path(path: Path) -> Path:
    """Get the module directory from a path"""
    # local import to avoid circular import
    from cognite_toolkit._cdf_tk.loaders import LOADER_BY_FOLDER_NAME

    if len(path.parts) == 1:
        raise ValueError("Path is not a module")

    for _ in range(len(path.parts)):
        if path.name in LOADER_BY_FOLDER_NAME and path.parent != path:
            return path.parent
        path = path.parent
    raise ValueError("Path is not part of a module")


def resource_folder_from_path(path: Path) -> str:
    """Get the resource_folder from a path"""
    # local import to avoid circular import
    from cognite_toolkit._cdf_tk.loaders import LOADER_BY_FOLDER_NAME

    for part in path.parts:
        if part in LOADER_BY_FOLDER_NAME:
            return part
    raise ValueError("Path does not contain a resource folder")


def is_module_path(path: Path) -> bool:
    """Check if a path is a module path"""
    # local import to avoid circular import
    from cognite_toolkit._cdf_tk.loaders import LOADER_BY_FOLDER_NAME

    if not path.is_dir():
        return False

    return any(sub_folder.name in LOADER_BY_FOLDER_NAME for sub_folder in path.iterdir() if sub_folder.is_dir())


def find_directory_with_subdirectories(
    directory_name: str | None, root_directory: Path
) -> tuple[Path | None, list[str]]:
    """Search for a directory with a specific name in the root_directory
    and return the directory and all subdirectories."""
    if directory_name is None:
        return None, []
    search = [root_directory]
    while search:
        current = search.pop()
        for root in current.iterdir():
            if not root.is_dir():
                continue
            if root.name == directory_name:
                return root, [d.name for d in root.iterdir() if d.is_dir()]
            search.append(root)
    return None, []


def parse_user_selected_modules(
    user_selected: list[str | Path] | None, organization_dir: Path | None = None
) -> list[str | Path]:
    """Parse user selected modules.

    The selected modules can be a mix of strings and Paths. If the string contains the module path separator, it is
    assumed to be a Path, otherwise we consider it a module name. If organization_dir is provided, we will convert
    relative paths to be relative to the organization_dir.

    Args:
        user_selected: The user selected modules
        organization_dir: The organization directory

    Returns:
        list[str | Path]: The parsed user selected modules

    """
    # The type of raw path is set just to make mypy happy.
    raw_paths = (selected for selected in user_selected or [] if isinstance(selected, Path))
    raw_str = (selected for selected in user_selected or [] if isinstance(selected, str))
    cleaned = (selected.replace("\\", "/") for selected in raw_str or [])
    all_selected: Iterable[str | Path] = itertools.chain(
        (Path(selected) if MODULE_PATH_SEP in selected else selected for selected in cleaned), raw_paths
    )

    if organization_dir:
        all_selected = (
            selected.relative_to(organization_dir)
            if isinstance(selected, Path) and selected.is_relative_to(organization_dir)
            else selected
            for selected in all_selected
        )

    return list(all_selected)



from __future__ import annotations

import shutil
import tempfile
from collections.abc import Iterable
from pathlib import Path
from typing import Any

import pytest
import yaml
from _pytest.mark import ParameterSet

from cognite_toolkit._cdf_tk.data_classes import BuildVariable, BuildVariables
from cognite_toolkit._cdf_tk.tk_warnings import (
    EnvironmentVariableMissingWarning,
    TemplateVariableWarning,
    catch_warnings,
)
from cognite_toolkit._cdf_tk.utils import (
    calculate_directory_hash,
    flatten_dict,
    iterate_modules,
    load_yaml_inject_variables,
    module_from_path,
    quote_int_value_by_key_in_yaml,
    stringify_value_by_key_in_yaml,
)
from cognite_toolkit._cdf_tk.utils.file import yaml_safe_dump
from cognite_toolkit._cdf_tk.utils.modules import module_directory_from_path
from cognite_toolkit._cdf_tk.validation import validate_modules_variables
from tests.data import DATA_FOLDER, PROJECT_FOR_TEST


class TestLoadYamlInjectVariables:
    def test_load_yaml_inject_variables(self, tmp_path: Path) -> None:
        my_file = tmp_path / "test.yaml"
        my_file.write_text(yaml_safe_dump({"test": "${TEST}"}))

        loaded = load_yaml_inject_variables(my_file, {"TEST": "my_injected_value"})

        assert loaded["test"] == "my_injected_value"

    def test_warning_when_missing_env_variable(self) -> None:
        path = Path("test.yaml")
        content = yaml_safe_dump({"test": "${TEST}"})
        expected_warning = EnvironmentVariableMissingWarning(path, frozenset({"TEST"}))

        with catch_warnings() as warning_list:
            load_yaml_inject_variables(content, {}, original_filepath=path)

        assert len(warning_list) == 1
        assert warning_list[0] == expected_warning


@pytest.mark.parametrize(
    "variable, expected_warnings",
    [
        pytest.param(
            BuildVariable("sourceId", "<change_me>", False, Path()),
            [TemplateVariableWarning(Path("config.yaml"), "<change_me>", "sourceId", "")],
            id="Single warning",
        ),
        pytest.param(
            BuildVariable("sourceId", "<change_me>", False, Path("a_module")),
            [TemplateVariableWarning(Path("config.yaml"), "<change_me>", "sourceId", "a_module")],
            id="Nested warning",
        ),
        pytest.param(
            BuildVariable("sourceId", "<change_me>", False, Path("a_super_module/a_module")),
            [TemplateVariableWarning(Path("config.yaml"), "<change_me>", "sourceId", "a_super_module.a_module")],
            id="Deep nested warning",
        ),
        pytest.param(BuildVariable("sourceId", "123", False, Path("a_module")), [], id="No warning"),
    ],
)
def test_validate_config_yaml(variable: BuildVariable, expected_warnings: list[TemplateVariableWarning]) -> None:
    warnings = validate_modules_variables(BuildVariables([variable]), Path("config.yaml"))

    assert sorted(warnings) == sorted(expected_warnings)


def test_calculate_hash_on_folder() -> None:
    folder = Path(DATA_FOLDER / "calc_hash_data")
    hash1 = calculate_directory_hash(folder)
    hash2 = calculate_directory_hash(folder)

    print(hash1)

    assert hash1 == "e60120ed03ebc1de314222a6a330dce08b7e2d77ec0929cd3c603cfdc08999ad", (
        f"The hash should not change as long as content in {folder} is not changed."
    )
    assert hash1 == hash2
    tempdir = Path(tempfile.mkdtemp())
    shutil.rmtree(tempdir)
    shutil.copytree(folder, tempdir)
    hash3 = calculate_directory_hash(tempdir)
    shutil.rmtree(tempdir)

    assert hash1 == hash3


def auth_variables_validate_test_cases():
    yield pytest.param(
        {
            "CDF_CLUSTER": "my_cluster",
            "CDF_PROJECT": "my_project",
            "LOGIN_FLOW": "token",
            "CDF_TOKEN": "12345",
        },
        False,
        "ok",
        [],
        {
            "cluster": "my_cluster",
            "project": "my_project",
            "cdf_url": "https://my_cluster.cognitedata.com",
            "login_flow": "token",
            "token": "12345",
            "client_id": None,
            "client_secret": None,
            "token_url": None,
            "tenant_id": None,
            "audience": "https://my_cluster.cognitedata.com",
            "scopes": "https://my_cluster.cognitedata.com/.default",
            "authority_url": None,
        },
        id="Happy path Token login",
    )

    yield pytest.param(
        {
            "CDF_CLUSTER": "my_cluster",
            "CDF_PROJECT": "my_project",
            "LOGIN_FLOW": "interactive",
            "IDP_CLIENT_ID": "7890",
            "IDP_TENANT_ID": "12345",
        },
        False,
        "ok",
        [],
        {
            "cluster": "my_cluster",
            "project": "my_project",
            "cdf_url": "https://my_cluster.cognitedata.com",
            "login_flow": "interactive",
            "token": None,
            "client_id": "7890",
            "client_secret": None,
            "token_url": "https://login.microsoftonline.com/12345/oauth2/v2.0/token",
            "tenant_id": "12345",
            "audience": "https://my_cluster.cognitedata.com",
            "scopes": "https://my_cluster.cognitedata.com/.default",
            "authority_url": "https://login.microsoftonline.com/12345",
        },
        id="Happy path Interactive login",
    )
    yield pytest.param(
        {
            "CDF_CLUSTER": "my_cluster",
            "CDF_PROJECT": "my_project",
            "LOGIN_FLOW": "client_credentials",
            "IDP_CLIENT_ID": "7890",
            "IDP_CLIENT_SECRET": "12345",
            "IDP_TENANT_ID": "12345",
        },
        False,
        "ok",
        [],
        {
            "cluster": "my_cluster",
            "project": "my_project",
            "cdf_url": "https://my_cluster.cognitedata.com",
            "login_flow": "client_credentials",
            "token": None,
            "client_id": "7890",
            "client_secret": "12345",
            "token_url": "https://login.microsoftonline.com/12345/oauth2/v2.0/token",
            "tenant_id": "12345",
            "audience": "https://my_cluster.cognitedata.com",
            "scopes": "https://my_cluster.cognitedata.com/.default",
            "authority_url": "https://login.microsoftonline.com/12345",
        },
        id="Happy path Client credentials login",
    )


class TestModuleFromPath:
    @pytest.mark.parametrize(
        "path, expected",
        [
            pytest.param(Path("cognite_modules/a_module/data_models/my_model.datamodel.yaml"), "a_module"),
            pytest.param(Path("cognite_modules/another_module/data_models/views/my_view.view.yaml"), "another_module"),
            pytest.param(
                Path("cognite_modules/parent_module/child_module/data_models/containers/my_container.container.yaml"),
                "child_module",
            ),
            pytest.param(
                Path("cognite_modules/parent_module/child_module/data_models/auth/my_group.group.yaml"), "child_module"
            ),
            pytest.param(Path("custom_modules/child_module/functions/functions.yaml"), "child_module"),
            pytest.param(Path("custom_modules/parent_module/child_module/functions/functions.yaml"), "child_module"),
        ],
    )
    def test_module_from_path(self, path: Path, expected: str):
        assert module_from_path(path) == expected


class TestIterateModules:
    def test_modules_project_for_tests(self):
        expected_modules = {
            PROJECT_FOR_TEST / "modules" / "a_module",
            PROJECT_FOR_TEST / "modules" / "another_module",
            PROJECT_FOR_TEST / "modules" / "parent_module" / "child_module",
        }

        actual_modules = {module for module, _ in iterate_modules(PROJECT_FOR_TEST)}

        assert actual_modules == expected_modules


@pytest.mark.parametrize(
    "input_, expected",
    [
        pytest.param({"a": {"b": 1, "c": 2}}, {("a", "b"): 1, ("a", "c"): 2}, id="Simple"),
        pytest.param({"a": {"b": {"c": 1}}}, {("a", "b", "c"): 1}, id="Nested"),
    ],
)
def test_flatten_dict(input_: dict[str, Any], expected: dict[str, Any]) -> None:
    actual = flatten_dict(input_)

    assert actual == expected


def quote_key_in_yaml_test_cases() -> Iterable[ParameterSet]:
    yield pytest.param(
        """space: my_space
externalID: myModel
version: 3_0_2""",
        '''space: my_space
externalID: myModel
version: "3_0_2"''',
        id="Single data model",
    )

    yield pytest.param(
        """- space: my_space
  externalId: myModel
  version: 1_000
- space: my_other_space
  externalId: myOtherModel
  version: 2_000
""",
        """- space: my_space
  externalId: myModel
  version: "1_000"
- space: my_other_space
  externalId: myOtherModel
  version: "2_000"
""",
        id="Two Data Models",
    )

    yield pytest.param(
        """space: my_space
externalID: myModel
version: '3_0_2'""",
        """space: my_space
externalID: myModel
version: '3_0_2'""",
        id="Single data model with single quoted version",
    )

    yield pytest.param(
        """- space: my_space
  externalId: myModel
  version: '1_000'
- space: my_other_space
  externalId: myOtherModel
  version: '2_000'
""",
        """- space: my_space
  externalId: myModel
  version: '1_000'
- space: my_other_space
  externalId: myOtherModel
  version: '2_000'
""",
        id="Two Data Models with single quoted version",
    )

    yield pytest.param(
        '''space: my_space
externalID: myModel
version: "3_0_2"''',
        '''space: my_space
externalID: myModel
version: "3_0_2"''',
        id="Single data model with double quoted version",
    )

    yield pytest.param(
        """- space: my_space
  externalId: myModel
  version: "1_000"
- space: my_other_space
  externalId: myOtherModel
  version: "2_000"
""",
        """- space: my_space
  externalId: myModel
  version: "1_000"
- space: my_other_space
  externalId: myOtherModel
  version: "2_000"
""",
        id="Two Data Models with double quoted version",
    )

    version_prop = """
externalId: CogniteSourceSystem
properties:
  version:
    container:
      externalId: CogniteSourceSystem
      space: sp_core_model
      type: container
    """
    yield pytest.param(
        version_prop,
        version_prop,
        id="Version property untouched",
    )


def stringify_value_by_key_in_yaml_test_cases() -> Iterable[ParameterSet]:
    yield pytest.param(
        """externalId: MyModel
config:
  data:
    debug: False
    runAll: False""",
        """externalId: MyModel
config: |
  data:
    debug: False
    runAll: False""",
        id="Stringify value under key 'config'",
    )
    input_ = """externalId: MyModel
config: |
  data:
   debug: False
   runAll: False"""
    yield pytest.param(input_, input_, id="Stringified value untouched")

    yield pytest.param(
        """config:
  data:
   debug: False
   runAll: False
externalId: MyModel""",
        """config: |
  data:
   debug: False
   runAll: False
externalId: MyModel""",
        id="Stringify value under key 'config' when it is not the first key",
    )

    input_ = """externalId: MyModel
config: another_string"""
    yield pytest.param(input_, input_, id="Stringified value untouched when it is not a dictionary")


class TestQuoteKeyInYAML:
    @pytest.mark.parametrize("raw, expected", list(quote_key_in_yaml_test_cases()))
    def test_quote_key_in_yaml(self, raw: str, expected: str) -> None:
        assert quote_int_value_by_key_in_yaml(raw, key="version") == expected

    @pytest.mark.parametrize("raw, expected", list(stringify_value_by_key_in_yaml_test_cases()))
    def test_stringify_value_by_key_in_yaml(self, raw: str, expected: str) -> None:
        actual = stringify_value_by_key_in_yaml(raw, key="config")
        assert actual == expected
        assert yaml.safe_load(actual) == yaml.safe_load(expected)


class TestModules:
    @pytest.mark.parametrize(
        "path, expected",
        [
            (Path("cdf_common/data_sets/demo.DataSet.yaml"), Path("cdf_common")),
            (Path("cdf_common/functions/contextualization_connection_writer"), Path("cdf_common")),
            (Path("sourcesystem/cdf_pi/auth/workflow.Group.yaml"), Path("sourcesystem/cdf_pi")),
        ],
    )
    def test_valid_module_directory_from_path(self, path: Path, expected: Path) -> None:
        assert module_directory_from_path(path) == expected



from pathlib import Path
from typing import Any
from unittest.mock import MagicMock

import pytest

from cognite_toolkit._cdf_tk.client.data_classes.raw import RawTable
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.data_classes import BuildEnvironment, BuiltResource, BuiltResourceList, SourceLocationEager
from cognite_toolkit._cdf_tk.loaders import RawFileLoader, RawTableLoader


class TestRawFileLoader:
    @pytest.mark.parametrize(
        "csv_content, expected_write",
        [
            (
                """myFloat,myInt,myString,myBool
0.1,1,hello,True
0.2,2,world,False
""",
                {
                    0: {"myFloat": 0.1, "myInt": 1, "myString": "hello", "myBool": True},
                    1: {"myFloat": 0.2, "myInt": 2, "myString": "world", "myBool": False},
                },
            )
        ],
    )
    def test_upload_dtypes(self, csv_content: str, expected_write: dict[int, Any]) -> None:
        with monkeypatch_toolkit_client() as client:
            loader = RawFileLoader.create_loader(client)
        csv_file = MagicMock(spec=Path)
        csv_file.read_bytes.return_value = csv_content.encode("utf-8")
        csv_file.exists.return_value = True
        csv_file.suffix = ".csv"
        source_file = MagicMock(spec=Path)
        source_file.with_suffix.return_value = csv_file

        state = BuildEnvironment()
        state.built_resources[RawFileLoader.folder_name] = BuiltResourceList(
            [
                BuiltResource(
                    RawTable("myDB", "myTable"),
                    SourceLocationEager(source_file, "1z234"),
                    RawTableLoader.kind,
                    None,
                    None,
                )
            ]
        )

        list(loader.upload(state, dry_run=False))

        assert client.raw.rows.insert_dataframe.call_count == 1
        _, kwargs = client.raw.rows.insert_dataframe.call_args
        written_to_cdf = kwargs["dataframe"].to_dict(orient="index")
        assert written_to_cdf == expected_write



from pydantic import Field

from .base import ToolkitResource


class DataSetYAML(ToolkitResource):
    external_id: str = Field(
        description="The external ID provided by the client.",
        max_length=255,
    )
    name: str | None = Field(
        default=None,
        description="The name of the data set.",
        min_length=1,
        max_length=50,
    )
    description: str | None = Field(
        default=None,
        description="The description of the data set.",
        min_length=1,
        max_length=500,
    )
    metadata: dict[str, str] | None = Field(
        default=None,
        description="Custom, application-specific metadata.",
    )
    write_protected: bool | None = Field(
        default=False,
        description="To write data to a write-protected data set.",
    )



from pathlib import Path
from typing import Annotated, Optional

import typer
from rich import print

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.commands import ModulesCommand, PullCommand
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from cognite_toolkit._version import __version__

CDF_TOML = CDFToml.load(Path.cwd())


class ModulesApp(typer.Typer):
    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.main)
        self.command()(self.init)
        self.command()(self.upgrade)
        self.command()(self.pull)
        self.command()(self.list)
        self.command()(self.add)

    def main(self, ctx: typer.Context) -> None:
        """Commands to manage modules"""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf modules --help[/] for more information.")

    def init(
        self,
        organization_dir: Annotated[
            Optional[Path],
            typer.Argument(
                help="Directory path to project to initialize or upgrade with templates.",
            ),
        ] = None,
        all: Annotated[
            bool,
            typer.Option(
                "--all",
                "-a",
                help="Copy all available templates.",
            ),
        ] = False,
        clean: Annotated[
            bool,
            typer.Option(
                "--clean",
                "-a",
                help="Clean target directory if it exists",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Initialize or upgrade a new CDF project with templates interactively."""

        cmd = ModulesCommand()
        cmd.run(
            lambda: cmd.init(
                organization_dir=organization_dir,
                select_all=all,
                clean=clean,
            )
        )

    def upgrade(
        self,
        organization_dir: Annotated[
            Path,
            typer.Option(
                "--organization-dir",
                "-o",
                help="Where to find the module templates to build from",
            ),
        ] = CDF_TOML.cdf.default_organization_dir,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Print details of each change applied in the upgrade process.",
            ),
        ] = False,
    ) -> None:
        cmd = ModulesCommand()
        cmd.run(lambda: cmd.upgrade(organization_dir=organization_dir, verbose=verbose))

    # This is a trick to use an f-string for the docstring
    upgrade.__doc__ = f"""Upgrade the existing CDF project modules to version {__version__}."""

    def add(
        self,
        organization_dir: Annotated[
            Path,
            typer.Option(
                "--organization-dir",
                "-o",
                help="Path to project directory with the modules. This is used to search for available functions.",
            ),
        ] = CDF_TOML.cdf.default_organization_dir,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Add one or more new module(s) to the project."""
        cmd = ModulesCommand()
        cmd.run(lambda: cmd.add(organization_dir=organization_dir))

    def pull(
        self,
        ctx: typer.Context,
        module_name_or_path: Annotated[
            Optional[str],
            typer.Argument(
                help="The module or path to module to pull from CDF. If not provided, interactive mode will be used.",
                allow_dash=True,
            ),
        ] = None,
        organization_dir: Annotated[
            Path,
            typer.Option(
                "--organization-dir",
                "-o",
                help="Where to find the module templates to build from",
            ),
        ] = CDF_TOML.cdf.default_organization_dir,
        build_env: Annotated[
            str,
            typer.Option(
                "--env",
                "-e",
                help="Build environment to use.",
            ),
        ] = CDF_TOML.cdf.default_env,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="Do now change the local files on the disk.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Print details of each change applied in the pull process.",
            ),
        ] = False,
    ) -> None:
        """Pull a module from CDF. This will overwrite the local files with the latest version from CDF."""
        cmd = PullCommand()
        env_vars = EnvironmentVariables.create_from_environment()
        cmd.run(
            lambda: cmd.pull_module(
                module_name_or_path=module_name_or_path,
                organization_dir=organization_dir,
                env=build_env,
                dry_run=dry_run,
                verbose=verbose,
                env_vars=env_vars,
            )
        )

    def list(
        self,
        organization_dir: Annotated[
            Path,
            typer.Option(
                "--organization-dir",
                "-o",
                help="Where to find the module templates to build from",
            ),
        ] = CDF_TOML.cdf.default_organization_dir,
        build_env: Annotated[
            Optional[str],
            typer.Option(
                "--env",
                help="Build environment to use.",
            ),
        ] = CDF_TOML.cdf.default_env,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """List all available modules in the project."""
        cmd = ModulesCommand()
        cmd.run(lambda: cmd.list(organization_dir=organization_dir, build_env_name=build_env))



from dataclasses import dataclass
from typing import ClassVar

from cognite.client.data_classes._base import CogniteResource
from cognite.client.utils._text import to_camel_case
from pydantic import BaseModel


class BaseModelResource(BaseModel, alias_generator=to_camel_case, extra="forbid"): ...


class ToolkitResource(BaseModelResource):
    _cdf_resource: ClassVar[type[CogniteResource]]

    def as_cognite_sdk_resource(self) -> CogniteResource:
        """Convert the model to a CDF resource."""
        return self._cdf_resource.load(self.model_dump(exclude_unset=True))


@dataclass(frozen=True)
class UnusedFieldWarning(Warning):
    """Warning for unused fields in the model."""

    field_names: frozenset[str]



from __future__ import annotations

import inspect
from typing import Any, Literal, get_args, get_origin

from .constants import ANY_INT, ANY_STR, ANYTHING, BASE_TYPES, CONTAINER_TYPES
from .data_classes import ParameterSet, ParameterSpec, ParameterSpecSet, ParameterValue
from .get_type_hints import _TypeHints
from .type_hint import TypeHint


def read_parameter_from_init_type_hints(cls_: type) -> ParameterSpecSet:
    return ParameterFromInitTypeHints().read(cls_)


class ParameterFromInitTypeHints:
    """Finds the parameters of a class by reading the type hints of the __init__ method."""

    def __init__(self) -> None:
        self.parameter_set = ParameterSpecSet()

    def read(self, cls_: type) -> ParameterSpecSet:
        self._read(cls_, tuple(), set())
        return self.parameter_set

    def _read(self, cls_: type, path: tuple[str | int, ...], seen: set[str]) -> None:
        if not hasattr(cls_, "__init__"):
            return None

        classes = _TypeHints.get_concrete_classes(cls_)
        try:
            seen.add(cls_.__name__)
        except AttributeError:
            # Python 3.9
            if str(cls_) == "typing.Any":
                return None
            raise
        seen.update(cls_.__name__ for cls_ in classes)
        for cls_ in classes:
            type_hints_by_name = _TypeHints.get_type_hints_by_name(cls_)
            for name, parameter in inspect.signature(cls_.__init__).parameters.items():  # type: ignore[misc]
                if name == "self" or parameter.kind in [parameter.VAR_POSITIONAL, parameter.VAR_KEYWORD]:
                    continue
                try:
                    hint = TypeHint(type_hints_by_name[name])
                except KeyError:
                    # Missing type hint
                    self.parameter_set.is_complete = False
                    continue

                is_required = parameter.default is inspect.Parameter.empty
                is_nullable = hint.is_nullable or parameter.default is None
                self.parameter_set.add(ParameterSpec((*path, name), hint.frozen_types, is_required, is_nullable))
                for subhint in hint.sub_hints:
                    if not subhint.is_base_type:
                        self._create_nested_parameters((name,), is_required, hint, path, seen)

    def _create_nested_parameters(
        self,
        parent_name: tuple[str | int, ...],
        is_parent_required: bool,
        hint: TypeHint,
        path: tuple[str | int, ...],
        seen: set[str],
    ) -> None:
        for sub_hint in hint.sub_hints:
            if sub_hint.is_dict_type:
                self._create_parameter_spec_dict(sub_hint, parent_name, path, seen)
            if sub_hint.is_list_type:
                self._create_parameter_spec_list(sub_hint, parent_name, is_parent_required, path, seen)
            if sub_hint.is_user_defined_class:
                self._create_parameter_spec_class(sub_hint, parent_name, is_parent_required, path, seen)
        return None

    def _create_parameter_spec_dict(
        self, hint: TypeHint, parent_name: tuple[str | int, ...], path: tuple[str | int, ...], seen: set[str]
    ) -> None:
        try:
            key, value = hint.container_args
        except ValueError:
            # There are no type hints for the dict
            self.parameter_set.add(
                ParameterSpec(
                    (*path, *parent_name, ANYTHING), frozenset({"unknown"}), is_required=False, _is_nullable=True
                )
            )
            return
        if get_origin(key) is Literal:
            args = get_args(key)
            key_path = args[0]
        elif key is str:
            key_path = ANY_STR
        else:
            raise NotImplementedError(
                f"Only Literal and str are supported for dict keys. {key.__name__} is not supported."
            )
        value_hint = TypeHint(value)

        self.parameter_set.add(
            ParameterSpec(
                (*path, *parent_name, key_path),
                value_hint.frozen_types,
                is_required=False,
                _is_nullable=value_hint.is_nullable,
            )
        )
        for sub_hint in value_hint.sub_hints:
            if not (sub_hint.is_base_type or sub_hint.is_any):
                self._create_nested_parameters(tuple(), False, sub_hint, (*path, *parent_name, ANY_STR), seen.copy())
        # if not (value_hint.is_base_type or value_hint.is_any):
        #     self._read(value, (*path, *parent_name, ANY_STR), seen.copy())

    def _create_parameter_spec_list(
        self,
        hint: TypeHint,
        parent_name: tuple[str | int, ...],
        parent_is_required: bool,
        path: tuple[str | int, ...],
        seen: set[str],
    ) -> None:
        try:
            item = hint.container_args[0]
        except IndexError:
            # There are no type hints for the list
            self.parameter_set.add(
                ParameterSpec(
                    (*path, *parent_name, ANYTHING), frozenset({"unknown"}), is_required=False, _is_nullable=True
                )
            )
            return
        item_hint = TypeHint(item)
        if item_hint.is_base_type:
            self.parameter_set.add(
                ParameterSpec(
                    (*path, *parent_name, ANY_INT),
                    item_hint.frozen_types,
                    is_required=False,
                    _is_nullable=item_hint.is_nullable,
                )
            )
        else:
            self._create_nested_parameters(
                tuple(), parent_is_required, item_hint, (*path, *parent_name, ANY_INT), seen.copy()
            )

    def _create_parameter_spec_class(
        self,
        hint: TypeHint,
        parent_name: tuple[str | int, ...],
        parent_is_required: bool,
        path: tuple[str | int, ...],
        seen: set[str],
    ) -> None:
        cls_ = hint.args[0]
        if cls_.__name__ in seen:
            # This is to avoid infinite recursion
            self.parameter_set.add(
                ParameterSpec(
                    (*path, *parent_name), frozenset({"dict"}), parent_is_required, _is_nullable=hint.is_nullable
                )
            )
        else:
            self._read(cls_, (*path, *parent_name), seen.copy())


def read_parameters_from_dict(raw: dict) -> ParameterSet[ParameterValue]:
    return _read_parameters_from_raw(raw, tuple())


def _read_parameters_from_raw(raw: dict | list | Any, path: tuple[str | int, ...]) -> ParameterSet[ParameterValue]:
    parameter_set = ParameterSet[ParameterValue]()
    if type(raw).__name__ in BASE_TYPES:
        parameter_set.add(ParameterValue(path, type(raw).__name__, raw))  # type: ignore[arg-type]
        return parameter_set
    if isinstance(raw, list):
        for i, item in enumerate(raw):
            parameter_set.update(_read_parameters_from_raw(item, (*path, i)))
        return parameter_set
    if isinstance(raw, dict):
        for key, value in raw.items():
            type_ = type(value).__name__
            if type_ in BASE_TYPES:
                parameter_set.add(ParameterValue((*path, key), type_, value))
            elif type_ in CONTAINER_TYPES:
                # We cannot include the value type for containers as it is not hashable
                parameter_set.add(ParameterValue((*path, key), type_, None))
            if isinstance(value, dict):
                parameter_set.update(_read_parameters_from_raw(value, (*path, key)))
            elif isinstance(value, list):
                for i, item in enumerate(value):
                    parameter_set.update(_read_parameters_from_raw(item, (*path, key, i)))
    return parameter_set



import pytest

from cognite_toolkit._cdf_tk.resource_classes import DatabaseYAML, TableYAML
from tests.test_unit.utils import find_resources


class TestRawDatabaseTableYAML:
    @pytest.mark.parametrize("data", list(find_resources("Database")))
    def test_load_valid_database(self, data: dict[str, object]) -> None:
        loaded = DatabaseYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data", list(find_resources("Table")))
    def test_load_valid_table(self, data: dict[str, object]) -> None:
        loaded = TableYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data
